"""
The ``kszx.desils_lrg`` module is based on data products from
Zhou et al 2023, "DESI LRG samples for cross-correlation",
https://arxiv.org/abs/2309.06443.

Data files are mostly found here:
  https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/.

With the exception of the randoms, which are here:
  https://data.desi.lbl.gov/public/ets/target/catalogs/dr9/0.49.0/randoms/
"""

import os
import yaml
import fitsio
import healpy
import warnings
import functools
import numpy as np

from . import Catalog
from . import io_utils
    

def read_galaxies(extended, download=False):
    r"""Reads LRGs with no imaging weights or quality cuts, and returns a Catalog object.

    After calling this function, you'll probably want to call :func:`~kszx.desils_lrg.compute_imaging_weights`
    and/or :func:`~kszx.desils_lrg.apply_quality_cuts`. You'll also want to call :func:`Catalog.apply_redshift_cut()`
    to retrict to an appropriate redshift range.
    
    Function arguments:

      - ``extended`` (boolean): determines whether "extended" or "main" catalog is read.
      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, with the following columns::
    
      ra_deg, dec_deg, z, zerr,  # sky location, redshift, redshift error
      pz_bin,                    # in {1,2,3,4}
      nobs_{g,r,z}, ebv, lrg_mask, maskbits      # for quality cuts
      photsys, galdepth_{g,r,z}, psfsize_{g,r,z} # for imaging weights
    """

    # Note: LRG catalog is split between a few FITS files (see _catalog_filename docstring).

    # First, the "main" FITS file.
    # cols = list of pairs (my_col_name, fits_col_name).
    cols = [ ('ra_deg','RA'), ('dec_deg','DEC'), ('z','Z_PHOT_MEDIAN') ]
    cols += [ (f'nobs_{x}', f'PIXEL_NOBS_{x.upper()}') for x in ['g','r','z'] ]
    cols += [ ('ebv','EBV'), ('lrg_mask','lrg_mask'), ('pz_bin','pz_bin') ]
    cols += [ ('maskbits','MASKBITS'), ('photsys','PHOTSYS') ]
    filename = _catalog_filename(extended, download=download, dlfunc='kszx.desils_lrg.read_galaxies')
    catalog = Catalog.from_fits(filename, cols)

    # Read zerr column (photo-z error) from the "pz" FITS file.
    filename = _catalog_filename(extended, suffix='pz', download=download, dlfunc='kszx.desils_lrg.read_galaxies')
    ctmp = Catalog.from_fits(filename, [('zerr','Z_PHOT_STD')])
    catalog.absorb_columns(ctmp, destructive=True)

    # Read GALDEPTH_{G,R,Z} and PSFSIZE_{G,R,Z} cols from the "more_2" FITS file.
    cols = [ (f'galdepth_{x}', f'GALDEPTH_{x.upper()}') for x in ['g','r','z'] ]
    cols += [ (f'psfsize_{x}', f'PSFSIZE_{x.upper()}') for x in ['g','r','z'] ]
    filename = _catalog_filename(extended, suffix='more_2', download=download, dlfunc='kszx.desils_lrg.read_galaxies')
    ctmp = Catalog.from_fits(filename, cols)
    catalog.absorb_columns(ctmp, destructive=True)

    return catalog


def read_randoms(ix_list, download=False):
    r"""Reads LRG random catalog with no quality cuts, and returns a Catalog object.

    After calling this function, you'll probably want to call :func:`~kszx.desils_lrg.apply_quality_cuts`.
    Note that DESILS-LRG randoms do not have redshifts!
    
    Function arguments:

      - ``ix_list`` (list): a list of integers $0 \le i < 200$. The DESILS-LRG randoms
        are split across 200 source files, and this argument determines which source files
        will be read.
    
      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Returns a :class:`kszx.Catalog` object, obtained by combining randoms from all
    source files in ``ix_list``, with the following columns::

      ra_deg, dec_deg,           # sky location but no redshift
      nobs_{g,r,z}, ebv, lrg_mask, maskbits  # for quality cuts
    """

    assert len(ix_list) >= 1
    assert len(set(ix_list)) == len(ix_list)  # no duplicates
    
    catalog_list = [ ]
    
    for ix in ix_list:
        assert 0 <= ix < 200   # generalize later

        # List of pairs (my_col_name, fits_col_name).
        cols = [ ('ra_deg','RA'), ('dec_deg','DEC'), ('ebv','EBV') ]
        cols += [ (f'nobs_{x}', f'NOBS_{x.upper()}') for x in ['g','r','z'] ]
        cols += [ ('maskbits','MASKBITS') ]
    
        filename = _random_filename(ix, mflag=False, download=download, dlfunc='kszx.desils_lrg.read_galaxies')
        catalog = Catalog.from_fits(filename, cols)

        # lrg_mask
        filename2 = _random_filename(ix, mflag=True, download=download, dlfunc='kszx.desils_lrg.read_galaxies')
        ctmp = Catalog.from_fits(filename2, [('lrg_mask','lrg_mask')])
        catalog.add_column('lrg_mask', ctmp.lrg_mask)

        catalog_list.append(catalog)

    return Catalog.concatenate(catalog_list, name='DESILS-LRG randoms', destructive=True)


def compute_imaging_weights(gcat, extended, ebv=True, download=False):
    r"""Adds a 'weights' column to the Catalog (usually called just after read_galaxies()).

    Note that this function can be called on the output of :func:`~kszx.desils_lrg.read_galaxies`,
    but not the output of :func:`~kszx.desils_lrg.read_randoms`.
    
    Function arguments:

      - ``gcat``: a :class:`~kszx.Catalog` object, obtained by calling
        :func:`kszx.desils_lrg.read_galaxies`.

      - ``extended`` (boolean): Slightly different imaging weights should be used for the
        "main" and "extended" DESILS-LRG samples, and this argument selects between them.
        (Note that ``extended`` is also an argument to :func:`~kszx.desils_lrg.read_galaxies`.)
    
      - ``ebv`` (boolean): The DESILS-LRG data products define two sets of imaging weights,
        which do or do not use the E(B-V) column. This argument selects between them.

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.
    
    Reference:
    
      https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/compute_imaging_weights.py

    Note: in addition to the ``weights`` column, this function also adds ``galdepth_gmag_ebv``,
    ``galdepth_rmag_ebv``, and ``galdepth_zmag_ebv`` columns to the Catalog (following the python
    script referenced above.)
    """

    # Weights array that will be added to Catalog at the end.
    weight = np.zeros(gcat.size)

    # Load weights (linear_coeffs yaml file)
    weights_path = _linear_coeffs_filename(extended, ebv, download, dlfunc='kszx.desils_lrg.read_galaxies')
    with open(weights_path, "r") as f:
        linear_coeffs = yaml.safe_load(f)

    # Convert depths to units of magnitude
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        gcat.add_column('galdepth_gmag_ebv', -2.5*(np.log10((5/np.sqrt(gcat.galdepth_g)))-9) - 3.214*gcat.ebv)
        gcat.add_column('galdepth_rmag_ebv', -2.5*(np.log10((5/np.sqrt(gcat.galdepth_r)))-9) - 2.165*gcat.ebv)
        gcat.add_column('galdepth_zmag_ebv', -2.5*(np.log10((5/np.sqrt(gcat.galdepth_z)))-9) - 1.211*gcat.ebv)

    # List of column names in linear regression ('EBV', 'PSFSIZE_G', 'galdepth_gmag_ebv', etc.)
    xnames_fit = list(linear_coeffs['south_bin_1'].keys())
    xnames_fit.remove('intercept')

    # Assign zero weights to objects with invalid imaging properties
    # (their fraction should be negligibly small)
    mask_bad = np.full(gcat.size, False)
    for xname in xnames_fit:
        col = getattr(gcat, xname.lower())
        mask_bad |= ~np.isfinite(col)
    print(f'kszx.desils_lrg.compute_imaging_mask: {np.sum(mask_bad)} invalid objects')
    
    for field in ['north', 'south']:
        photsys = field[0].upper()   # 'N' or 'S'

        for bin_index in range(1, 5):  # 4 bins
            mask_bin = (gcat.photsys == photsys)
            mask_bin &= (gcat.pz_bin == bin_index)
            mask_bin &= ~mask_bad

            # wt = intercept + sum(coeff * col)
            bin_str = f'{field}_bin_{bin_index}'
            wt = linear_coeffs[bin_str]['intercept']
            
            for xname in xnames_fit:
                coeff = linear_coeffs[bin_str][xname]
                col = getattr(gcat, xname.lower())
                wt += coeff * col[mask_bin]

            weight[mask_bin] = 1.0 / wt    # 1/predicted_density as weights for objects

    gcat.add_column('weight', weight)


def apply_quality_cuts(catalog, min_nobs=2, max_ebv=0.15, max_stardens=2500, lrg_mask=True, maskbits=True, island_mask=True, mask_negative_zerr=True, download=True):
    r"""Applies LRG quality cuts in-place to a catalog (usually called just after read_galaxies() or read_randoms()).

    Function arguments:

      - ``catalog``: a :class:`~kszx.Catalog` object, obtained by calling
        :func:`kszx.desils_lrg.read_galaxies` or :func:`kszx.desils_lrg.read_randoms`.

      - ``min_nobs`` (integer): min allowed value for NOBS_G, NOBS_R, NOBS_Z.
        Default value (2) comes from the references below.

      - ``max_ebv`` (float): max allowed value for E(B-V).
        Default value (0.15) comes from the references below.

      - ``max_stardens`` (float): max allowed stellar density (in deg^{-2})).
        Note that stellar density isn't in the catalogs, and will be inferred from an external
        data product. Default value (2500) comes from the references below.
    
      - ``lrg_mask`` (boolean): whether to apply 'lrg_mask' column from catalog.
        (See third reference below.)

      - ``maskbits`` (boolean): whether to apply DESILS mask bits 1 (BRIGHT), 12 (GALAXY),
        and 13 (CLUSTER). This is strictly weaker than ``lrg_mask``. (See fourth reference
        below.)
    
      - ``island_mask`` (boolean): If True, then some islands in the NGC will be masked.

      - ``mask_negative_zerr`` (boolean): If True, galaxies with photo-z error < 0
        will be masked. (This flag was added by Kendrick, and is not in Rongpu's original
        script. Note that there are ~3000 such galaxies in the extended catalog, and only
        2 of them pass the other qualtiy cuts.)

      - ``download`` (boolean): if True, then all needed data files will be auto-downloaded.

    Reference:
    
       - https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/quality_cuts.py
       - https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/randoms_quality_cuts.py
       - https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/lrgmask_v1.1/README.txt
       - https://www.legacysurvey.org/dr10/bitmasks/
    """

    if island_mask:
        mask = ~((catalog.dec_deg < -10.5) & (catalog.ra_deg > 120) & (catalog.ra_deg < 260))
        catalog.apply_boolean_mask(mask, name = 'DESILS-LRG quality cut: remove NGC islands')

    if (min_nobs is not None) and (min_nobs > 0):
        mask = (catalog.nobs_g >= min_nobs) & (catalog.nobs_r >= min_nobs) & (catalog.nobs_z >= min_nobs)
        catalog.apply_boolean_mask(mask, name = f'DESILS-LRG quality cut: nobs >= {min_nobs}')

    if lrg_mask:
        bits = (1<<1) | (1<<12) | (1<<13)
        mask = np.logical_and(catalog.lrg_mask == 0, (catalog.maskbits & bits) == 0)
        catalog.apply_boolean_mask(mask, name = 'DESILS-LRG quality cut: LRG maskbits == 0')

    if max_ebv is not None:
        mask = (catalog.ebv < max_ebv)
        catalog.apply_boolean_mask(mask, name = f'DESILS-LRG quality cut: EBV < {max_ebv}')

    if max_stardens is not None:
        m = read_stardens_map(download=download)
        nside = healpy.npix2nside(len(m))
        ix = healpy.ang2pix(nside, catalog.ra_deg, catalog.dec_deg, lonlat=True)
        stardens = m[ix]
        mask = (stardens < max_stardens)
        catalog.apply_boolean_mask(mask, name = f'DESILS-LRG quality cut: stardens < {max_stardens}')

    if mask_negative_zerr and ('zerr' in catalog.col_names):
        mask = (catalog.zerr >= 0)
        catalog.apply_boolean_mask(mask, name = f'DESILS-LRG quality cut: zerr >= 0')
    

@functools.cache
def read_stardens_map(download=False):
    """Returns nside=64 ring-ordered healpix map, cached between calls to read_stardens_map().

    Intended as a helper for apply_quality_cuts(), but may be independently useful."""
    
    filename = _desils_lrg_path('misc/pixweight-dr7.1-0.22.0_stardens_64_ring.fits', download=True, dlfunc='kszx.desils_lrg.read_galaxies')
    
    print(f'Reading {filename}\n', end='')
    f = fitsio.read(filename)

    nside = 64
    npix = 12 * nside**2
    
    assert np.all(f['HPXPIXEL'] == np.arange(npix))
    return np.asarray(f['STARDENS'], dtype=np.float32)  # convert big-endian float32 -> native


####################################################################################################


def _catalog_filename(extended, suffix=None, download=False, dlfunc=None):
    """The 'extended' argument should be True or False.
    
    Catalogs are split across multiple FITS files, distinguished by 'suffix':

        suffix=None      TARGETID, RA, DEC, EBV, PIXEL_NOBS_{G,R,Z}, MASKBITS, 
                         PHOTSYS, Z_PHOT_MEDIAN, lrg_mask, pz_bin

        suffix='more_1'  GAIA_PHOT_{BP,RP}_MEAN_MAG, GAIA_ASTROMETRIC_EXCESS_NOISE,
                         FITBITS, FRACFLUX_{G,R,Z,W1,W2}, FRACMASKED_{G,R,Z},
                         FRACIN_{G,R,Z}, FIBERTOTFLUX_G, SHAPE_{R,R_IVAR},
                         SHAPE_{E1,E2}, SERSIC, DCHISSQ

        suffix='more_2'  GALDEPTH_{G,R,Z}, PSFSIZE_{G,R,Z,W1,W2}, PSFSIZE_{G,R,Z}

        suffix='photom'  TYPE, EBC, FLUX_{G,R,Z,W1,W2}, FLUX_IVAR_{G,R,Z,W1,W2},
                         MW_TRANSMISSION_{G,R,Z,W1,W2}, FIBERFLUX_{G,R,Z},
                         FIBERTOTFLUX_{R,Z}, GAIA_PHOT_G_MEAN_MAG

        suffix='pz'      Z_PHOT_{MEAN,MEDIAN,STD,L68,U68,L95,U95}, 
                         Z_SPEC, SURVEY, TRAINING

        suffix='pzbins-weights'          weight
        suffix='pzbins-weights_no_ebv'   weight

    For descriptions of some of these columns, see:

        https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/README.txt        
        https://www.legacysurvey.org/dr9/files/#sweep-catalogs-region-sweep
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """

    lrg = 'extended_lrg' if extended else 'lrg'
    slist = [ None, 'more_1', 'more_2', 'photom', 'pz', 'pzbins-weights', 'pzbins-weights_no_ebv' ]
    
    if suffix not in slist:
        raise RuntimeError(f"kszx.desils_lrg._catalog_filename: suffix='{suffix}' invalid."
                           + f" Valid suffixes are: {slist}")

    if suffix is None:
        return _desils_lrg_path(f'catalogs/dr9_{lrg}_pzbins.fits', download=download, dlfunc=dlfunc)
    else:
        return _desils_lrg_path(f'catalogs/more/dr9_{lrg}_{suffix}.fits', download=download, dlfunc=dlfunc)


def _random_filename(ix, mflag, download=False, dlfunc=None):
    """The index 'ix' satisfies 0 <= ix < 200.

    The randoms are split across two FITS files, distinguished through the boolean 'mflag' arg:
      - mflag=False   ETS file containing RA, DEC, NOBS_{G,R,Z}, EBV  (+ more cols I don't use)
      - mflag=True    LRG mask file containing boolean 'lrg_mask' col
    """

    assert 0 <= ix < 200
    s = f'{(ix//20)+1}-{(ix%20)}'   # E.g. (ix=31) -> '2-11'

    if mflag:
        return _desils_lrg_path(f'catalogs/lrgmask_v1.1/randoms-{s}-lrgmask_v1.1.fits.gz', download=download, dlfunc=dlfunc)
    else:
        return _desils_ets_path(f'target/catalogs/dr9/0.49.0/randoms/resolve/randoms-{s}.fits', download=download, dlfunc=dlfunc)


def _linear_coeffs_filename(extended, ebv, download=False, dlfunc=None):
    """The 'extended' and 'ebv' arguments should be True or False."""

    prefix = 'extended' if extended else 'main'
    suffix = '' if ebv else '_no_ebv'
    return _desils_lrg_path(f'catalogs/imaging_weights/{prefix}_lrg_linear_coeffs_pz{suffix}.yaml', download=download, dlfunc=dlfunc)

    
def _desils_lrg_path(relpath, download=False, dlfunc=None):
    """Example: _desils_lrs_path('catalogs/dr9_lrg_pzbins.fits').
        -> local filename /data/desils/lrg_xcorr_2023/v1/catalogs/dr9_lrg_pzbins.fits
        -> url https://data.desi.lbl.gov/public/papers/c3/lrg_xcorr_2023/v1/catalogs/dr9_lrg_pzbins.fits

    Intended to be called through wrapper, e.g. _catalog_filename()."""
    
    relpath = os.path.join('lrg_xcorr_2023/v1', relpath)
    desils_base_dir = os.path.join(io_utils.get_data_dir(), 'desils')
    abspath = os.path.join(desils_base_dir, relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://data.desi.lbl.gov/public/papers/c3/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
    
    return abspath


def _desils_ets_path(relpath, download=False, dlfunc=None):
    """Example: _desils_ets_path('target/catalogs/dr9/0.49.0/randoms/resolve/randoms-1-0.fits')
        -> local filename /data/desils/ets/target/catalogs/dr9/0.49.0/randoms/resolve/randoms-1-0.fits
        -> url https://data.desi.lbl.gov/public/ets/target/catalogs/dr9/0.49.0/randoms/resolve/randoms-1-0.fits

    Intended to be called through wrapper, e.g. _random_filename()."""

    relpath = os.path.join('ets', relpath)
    desils_base_dir = os.path.join(io_utils.get_data_dir(), 'desils')
    abspath = os.path.join(desils_base_dir, relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://data.desi.lbl.gov/public/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
    
    return abspath



import io    # StringIO
import functools
import numpy as np


class Box:
    def __init__(self, npix, pixsize, cpos=None):
        r"""The Box class represents a pixelization of an N-dimensional box.

        The Box class requires the caller to manually choose the box size and observer
        location. See :class:`~kszx.BoundingBox` for a subclass which automatically 
        chooses these parameters to match a galaxy survey.

        Constructor args:

            - npix (tuple or 1-d array): 
               Grid dimensions, e.g. (1024,1024,1024)
               For a non-cubic box, use unequal dimensions, e.g. (512,512,1024).

            - pixsize (float): 
               Pixel side length (caller-specified units, usually Mpc).
               Note that pixsize is a scalar, i.e. pixels are always cubic.

            - cpos (tuple, 1-d array, or None):
               Location of box center, in coordinate system with observer at origin.

               Defaults to cpos=(npix*pixsize)/2, i.e. observer at lower-left corner.
               For a box with observer at the center, use cpos=(0,0,0).

               The observer location matters in some situations (e.g. gridding a galaxy 
               catalog, where galaxy locations are specified in observer coordinates),
               but not others (e.g. taking an FFT).

        Maps and array shapes:

             Note that we define a Box class, but not a Map class.

             Instead, a real-space map is represented by a pair (box, arr), where 'arr' is a 
             numpy array with dtype=float. Similarly, a Fourier-space map is represented by a 
             pair (box, arr), where 'arr' is a numpy array with dtype=complex.

             The real-space and Fourier-space array shapes are related as follows:

        $$(\mbox{real-space shape}) = (n_0, n_1, \cdots, n_{d-1})$$
        $$(\mbox{Fourier-space shape})= (n_0, n_1, \cdots, \lfloor n_{d-1}/2 \rfloor + 1)$$

        Members:
        
           - ndim (integer): Number of dimensions (usually 3).
           - pixsize (float): Pixel side length in user-defined coordinates.
           - boxsize (length-ndim array): Box side lengths in user-defined coordinates (equal to npix * pixsize).
           - npix (length-ndim array): Real-space map shape, represented as numpy array.
           - real_space_shape (tuple): same as 'npix', but represented as tuple instead of 1-d array
           - nk (length-ndim array): Fourier-space map shape, represented as numpy array.
           - fourier_space_shape (tuple) same as 'nk', but represented as tuple instead of 1-d array
           - cpos (length-ndim array): location of box center, in observer coordinates
           - lpos (length-ndim array): location of lower left box corner, in observer coordinates
           - rpos (length-ndim array): location of upper right box corner, in observer coordinates
           - kfund (length-ndim array): Lowest frequency on each axis (equal to 2pi/boxsize)
           - knyq (float): Nyquist frequency (equal to pi/pixsize)
           - box_volume (float): Box volume (equal to prod(boxsize))
           - pixel_volume (float): Pixel volume (equal to pixsize^N)

        Fourier conventions:

        $$f(k) = V_{pix} \sum_x f(x) e^{-ik\cdot x}$$

        $$f(x) = V_{box}^{-1} \sum_k f(k) e^{ik\cdot x}$$

        $$\langle f(k) f(k')^* \rangle = V_{box} P(k) \delta_{k,k'}$$
        """
        
        npix = np.asarray(npix)
        assert npix.ndim == 1
        assert npix.dtype.kind in ['i','u']    # integer-valued
        assert np.all(npix > 0)

        ndim = len(npix)
        assert 1 <= ndim <= 3

        pixsize = float(pixsize)
        assert(pixsize > 0)
        
        if cpos is not None:
            cpos = np.array(cpos, dtype=float)
            assert cpos.shape == npix.shape

        # Real-space shape
        self.ndim = ndim
        self.npix = np.array(npix, dtype=int)
        self.real_space_shape = tuple(self.npix)

        # Fourier-space shape
        self.nk = np.copy(self.npix)
        self.nk[-1] = (self.nk[-1] // 2) + 1
        self.fourier_space_shape = tuple(self.nk)
 
        # Lengths / volumes / wavenumbers
        self.pixsize = pixsize
        self.boxsize = self.npix * self.pixsize
        self.kfund = 2*np.pi / self.boxsize
        self.knyq = np.pi / self.pixsize
        self.box_volume = np.prod(self.boxsize)
        self.pixel_volume = (self.pixsize)**ndim

        # Box location in observer coordinates.
        t = 0.5 * (self.npix - 1) * self.pixsize
        self.cpos = cpos if (cpos is not None) else (self.boxsize/2.)
        self.lpos = self.cpos - t
        self.rpos = self.cpos + t

        # Used internally in self.get_r(): which real-space pixel has the smallest r-coordinate?
        self._ix_smallest_r = -self.lpos / self.pixsize
        self._ix_smallest_r = np.array(np.round(self._ix_smallest_r), dtype=int)
        self._ix_smallest_r = np.maximum(self._ix_smallest_r, 0)
        self._ix_smallest_r = np.minimum(self._ix_smallest_r, self.npix-1)
        self._ix_smallest_r = tuple(self._ix_smallest_r)

        
    def is_real_space_map(self, arr):
        """Returns True if array 'arr' has the right shape/dtype for real-space map."""
        return isinstance(arr, np.ndarray) and np.all(arr.shape == self.real_space_shape) and (arr.dtype == float)

    
    def is_fourier_space_map(self, arr):
        """Returns True if array 'arr' has the right shape/dtype for Fourier-space map."""
        return isinstance(arr, np.ndarray) and np.all(arr.shape == self.fourier_space_shape) and (arr.dtype == complex)


    def get_k_component(self, axis, zero_nyquist=True, one_dimensional=False):
        """Returns array (k[0], k[1], ...) containing signed k-values along specified axis.

        Args:
            axis (integer): satisfies 0 <= axis < ndim.
            zero_nyquist (boolean): if True (recommended) zero will be returned at k_nyquist.
            one_dimensonal (booolean):
               - if True, the returned array will have shape (N,), where N = self.nk[axis].
               - if False, the returned array will have shape (1,...,1, N, 1,...1).

        Returns: real-valued array with shape described above.

        Notes:
            - k-values include the usual factor (2pi / boxsize).
            - Roughly half the k-values will be positive, and roughly half will be
              negative (except for the last axis, where negative values are not stored).
        """

        nd = self.ndim
        assert 0 <= axis < nd

        n = self.npix[axis]
        nk = self.nk[axis]
        ret = np.arange(nk, dtype=float)

        if axis != (self.ndim-1):
            m = n//2 + 1
            ret[m:] = np.arange(m-n, 0, dtype=float)

        if zero_nyquist and (n % 2) == 0:
            ret[n//2] = 0

        ret *= self.kfund[axis]
        
        if not one_dimensional:
            shape = (1,)*axis + (nk,) + (1,)*(nd-axis-1)
            ret = ret.reshape(shape)

        return ret


    def get_k(self, exponent=1, regulate=False):
        """Returns an N-dimensional array containing |k|^exponent values.

        Args:
            exponent (float): return |k|^exponent. If k<0, then regulate=True is implied.
            regulate (boolean): if True, then replace k=0 by k=(2pi/boxsize).

        Returns: real-valued array with shape (self.fourier_space_shape).
        """

        k2 = [ self.get_k_component(axis, zero_nyquist=False)**2 for axis in range(self.ndim) ]
        ret = functools.reduce(np.add, k2)
            
        if regulate or (exponent < 0):
            ret[(0,)*self.ndim] = np.min(self.kfund)**2

        if exponent != 2:
            ret **= (0.5*exponent)

        return ret


    def get_r_component(self, axis, one_dimensional=False):
        """Returns array (r[0], r[1], ...) containing signed observer coordinate along specified axis.

        Args:
            axis (integer): satisfies 0 <= axis < ndim.
            one_dimensonal (booolean):
               - if True, the returned array will have shape (N,), where N = self.npix[axis].
               - if False, the returned array will have shape (1,...,1, N, 1,...1).

        Returns: real-valued array with shape described above.
        
        Note: returned r-components are in "observer" coordinates, i.e. the observer is at
        the origin, and the box corners are given by self.lpos, self.rpos.
        """        
        
        assert 0 <= axis < self.ndim

        n = self.npix[axis]
        ret = np.arange(n,dtype=float) - 0.5*(n-1)
        ret = (ret * self.pixsize) + self.cpos[axis]

        if not one_dimensional:
            shape = (1,)*axis + (n,) + (1,)*(self.ndim-axis-1)
            ret = ret.reshape(shape)

        return ret


    def get_r(self, exponent=1, regulate=False, eps=1.0e-6):
        """Returns an N-dimensional array containing |r|^exponent values.

        Args:
            exponent (float): return |r|^exponent. If k<0, then regulate=True is implied.
            regulate (boolean): if True, then replace r = max(r,eps*pixsize).
            eps (float): only used if regulation is performed.

        Returns: real-valued array with shape (self.real_space_shape).
        
        Note: returned r-values are in "observer" coordinates, i.e. the observer is at
        the origin, and the box corners are given by self.lpos, self.rpos.
        """
        
        assert 0 < eps < 0.5

        r2 = [ self.get_r_component(axis)**2 for axis in range(self.ndim) ]
        ret = functools.reduce(np.add, r2)
        
        if regulate or (exponent < 0):
            i = self._ix_smallest_r
            ret[i] = max(ret[i], (eps * self.pixsize)**2)

        if exponent != 2:
            ret **= (0.5 * exponent)

        return ret


    def _print_box_members(self, f, end=','):
        """Helper function called by __str__(). (This factorization is convenient for the BoundingBox subclass.)"""        
        print(f'    npix = {self.npix},', file=f)
        print(f'    pixsize = {self.pixsize},', file=f)
        print(f'    boxsize = {self.boxsize},', file=f)
        print(f'    kfund = {self.kfund},', file=f)
        print(f'    knyq = {self.knyq},', file=f)
        print(f'    lpos = {self.lpos},', file=f)
        print(f'    cpos = {self.cpos},', file=f)
        print(f'    rpos = {self.rpos}{end}', file=f)
        
    def __str__(self):
        with io.StringIO() as f:
            print(f'Box(', file=f)
            self._print_box_members(f)
            print(f')', file=f)
            return f.getvalue()
    
    def __repr__(self):
        return str(self)



from . import test_box
from . import test_lss

def run_all_tests():
    test_box.test_k_component()
    test_box.test_r_component()
    test_box.test_smallest_r()
    test_lss.test_fft_inverses()
    test_lss.test_fft_transposes()
    test_lss.test_interpolation()
    test_lss.test_interpolation_gridding_consistency()
    test_lss.test_simulate_gaussian()
    test_lss.test_estimate_power_spectrum()
    test_lss.test_kbin_average()

    #test_lss.monte_carlo_simulate_gaussian([4,6,1], 10.0)
    #test_lss.monte_carlo_simulate_gaussian([5,4,6], 10.0)
    #test_lss.monte_carlo_simulate_gaussian([6,7,4], 10.0)
    #test_lss.monte_carlo_simulate_gaussian([4,4,5], 10.0)



import healpy
import warnings
import numpy as np

from . import utils
from . import io_utils

from .Catalog import Catalog


def read_map(filename):
    print(f'Reading {filename}\n', end='')
    
    # Suppress superfluous warnings from healpy.fitsfunc.read_map():
    #   .../healpy/fitsfunc.py:391: UserWarning: NSIDE = 4
    #   .../healpy/fitsfunc.py:400: UserWarning: ORDERING = RING in fits file
    #   .../healpy/fitsfunc.py:428: UserWarning: INDXSCHM = IMPLICIT

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        m = healpy.read_map(filename, dtype=None, nest=False)
        
    # Convert from big-endian float (numpy dtype '>f4') to native float
    return np.asarray(m, float)


def write_map(filename, m):
    print(f'Writing {filename}\n', end='')
    io_utils.mkdir_containing(filename)
        
    # Supply (dtype=m.dtype, overwrite=True).
    healpy.write_map(filename, m, dtype=m.dtype, overwrite=True)


def map_from_catalog(nside, gcat, weights=None, rcat=None, rweights=None, normalized=True):
    """Returns healpix map (float array with length 12*nside^2).

    The (rcat, rweights) arguments represent "randoms" to be subtracted.
    NOTE: the rweights are renormalized so that sum(rweights) = -sum(weights)!

    If normalized=True, then the output map includes a factor 1 / (pixel area).
    This normalization best represents a sum of delta functions f(x) = sum_j w_j delta^2(th_j).
    """
    
    assert isinstance(gcat, Catalog)
    assert isinstance(rcat, Catalog) or (rcat is None)

    if (rcat is None) and (rweights is not None):
        raise RuntimeError("kszx.healpix_utils.map_from_catalog(): 'rcat' arg is None, but 'rweights' arg is not None")

    weights = utils.asarray(weights, 'kszx.healpix_utils.map_from_catalog', 'weights', dtype=float, allow_none=True)
    rweights = utils.asarray(rweights, 'kszx.healpix_utils.map_from_catalog', 'rweights', dtype=float, allow_none=True)

    assert (weights is None) or (weights.ndim == 0) or (weights.shape == (gcat.size,))
    assert (rweights is None) or (rweights.ndim == 0) or (rweights.shape == (rcat.size,))
    
    npix = healpy.nside2npix(nside)
    ret = np.zeros(npix)

    ipix = healpy.ang2pix(nside, gcat.ra_deg, gcat.dec_deg, lonlat=True)
    wsum = utils.scattered_add(ret, ipix, weights)

    if rcat is not None:
        ipix = healpy.ang2pix(nside, rcat.ra_deg, rcat.dec_deg, lonlat=True)
        utils.scattered_add(ret, ipix, rweights, normalize_sum = -wsum)

    if normalized:
        ret /= healpy.nside2pixarea(nside)
    
    return ret


####################################################################################################


def read_alm(filename):
    print(f'Reading {filename}\n', end='')

    alm, mmax = healpy.read_alm(filename, return_mmax=True)
    expected_nelts = ((mmax+1) * (mmax+2)) // 2

    if alm.shape != (expected_nelts,):
        raise RuntimeError(f'{filename}: expected lmax==mmax')

    return alm


def lmax_of_alm(alm):
    assert alm.ndim == 1
    
    s = alm.size
    lmax = int((2*s)**0.5) - 1
    
    if (s != ((lmax+1)*(lmax+2))//2) or (lmax < 0):
        raise RuntimeError(f'{alm.size=} is not of the form (lmax+1)(lmax+2)/2')
    
    return lmax


def degrade_alm(alm, dst_lmax):
    """I'm surprised this function isn't in healpy or pixell!"""
    
    src_lmax = lmax_of_alm(alm)
    assert 0 <= dst_lmax <= src_lmax

    ret = np.zeros(((dst_lmax+1)*(dst_lmax+2)) // 2, dtype=complex)
    d = 0
    s = 0
    
    for m in range(dst_lmax+1):
        dn = dst_lmax - m + 1
        sn = src_lmax - m + 1
        ret[d:(d+dn)] = alm[s:(s+dn)]
        d += dn
        s += sn

    return ret


def plot_alm(alm, nside=128, fwhm_deg=None, lmax=None):
    src_lmax = lmax_of_alm(alm)
    
    if (lmax is not None) and (lmax < src_lmax):
        alm = degrade_alm(alm, lmax)
        src_lmax = lmax

    if fwhm_deg is not None:
        assert fwhm_deg >= 0.0
        b = (fwhm_deg * np.pi/180.)**2 / (16. * np.log(2.))
        bl = np.exp(-b * np.arange(src_lmax+1)**2)
        alm = healpy.almxfl(alm, bl)

    m = healpy.alm2map(alm, nside)
    healpy.mollview(m)



import numpy as np
import scipy.special
import kszx.utils as utils


class RegulatedDeconvolver:
    def __init__(self, zobs_vec, zerr_vec, zbin_width, soft_zmax=None):
        """Given the observed 2-d (zobs,zerr) distribution, models the 3-d (ztrue,zobs,zerr) distribution.

        zobs_vec: array of shape (ngal,) containing observed (photometric) redshifts.

        zerr_vec: array of shape (ngal,) containing estimated photo-z errors.

        zbin_width (scalar): used internally for binning. (Recommend ~0.01 for DESILS-LRG.)

        soft_zmax (scalar): used internally; galaxies with (zobs > soft_zmax) are discarded.
        Should be chosen significantly larger than the max redshift of interest.
        
        (For DESILS-LRG with zmax=1, recommend soft_zmax=1.5. If this parameter is omitted,
        then you'll get a tail of rare galaxies with z~5, which is annoying.)
        """
        
        assert zobs_vec.ndim == 1
        assert zobs_vec.shape == zerr_vec.shape
        assert np.all(zobs_vec >= 0)
        assert np.all(zerr_vec > 0)

        if soft_zmax is not None:
            mask = (zobs_vec <= soft_zmax)
            zobs_vec = zobs_vec[mask]
            zerr_vec = zerr_vec[mask]
        
        self.zobs_vec = zobs_vec
        self.zerr_vec = zerr_vec
        self.zbin_width = zbin_width
        self.zmin = np.min(zobs_vec)
        self.zmax = np.max(zobs_vec)
        self.soft_zmax = soft_zmax

        self.nzbins = np.array(self.zmax/zbin_width, dtype=int)
        self.zbin_edges = np.arange(self.nzbins+1) * zbin_width
        
        # The variable y = log(zerr + zbin_width) is used to define zerr bins.
        y = np.log(zerr_vec + self.zbin_width)
        self.ymin = np.min(y) - 1.0e-8
        self.ymax = np.max(y) + 1.0e-8
        self.nybins = int((self.ymax - self.ymin) / 0.2) + 1
        self.ybin_width = (self.ymax - self.ymin) / self.nybins
        
        iz = np.array(zobs_vec / zbin_width, dtype=int)
        iz = np.clip(iz, 0, self.nzbins-1)

        iy = np.array((y - self.ymin) / self.ybin_width, dtype=int)
        iy = np.clip(iy, 0, self.nybins-1)

        # self.pobs[iy,iz] is an array of shape (self.nybins, self.nzbins).
        self.pobs = np.bincount(iy*self.nzbins + iz, minlength=self.nybins*self.nzbins)
        self.pobs = np.reshape(self.pobs, (self.nybins, self.nzbins))
        self.pobs += 1   # regulate
        self.pobs = self.pobs / np.sum(self.pobs,axis=1).reshape((-1,1))  # normalize to PDF in z-bins (for each y-bin)


    def _sample(self, zobs, zerr):
        """Helper for sample(): Sample from conditional distribution P(ztrue|zobs,zerr)."""
        
        n = len(zobs)
        nb = self.nzbins
        rng = np.random.default_rng()

        # iy, iz = bin indices (shape (n,))
        # Reminder: the variable y = log(zerr + zbin_width) is used to define zerr bins.
        y = np.log(zerr + self.zbin_width)
        iy = np.array((y - self.ymin) / self.ybin_width, dtype=int)
        iy = np.clip(iy, 0, self.nybins-1)
        
        # pobs = shape(n,nb)
        pobs = self.pobs[iy]
        assert pobs.shape == (n,nb)

        # cdf = shape (n,nb+1), not "reweighted by pobs".
        cdf = self.C((self.zbin_edges.reshape((1,-1)) - zobs.reshape((-1,1))) / zerr.reshape((-1,1)))
        assert cdf.shape == (n,nb+1)
        
        # pdf = shape (n,nb), "reweighted by pobs"
        pdf = cdf[:,1:] - cdf[:,:-1]
        pdf *= pobs
        pdf /= np.sum(pdf,axis=1).reshape((-1,1))

        # ztrue_bins has shape (n,). I don't think there's any way to vectorize this loop.
        ztrue_bins = np.array([ rng.choice(nb, p=pdf[i,:]) for i in range(n) ])
        
        cdf_lo = cdf[np.arange(n), ztrue_bins]
        cdf_hi = cdf[np.arange(n), ztrue_bins+1]
        c = cdf_lo + (cdf_hi-cdf_lo) * rng.uniform(size=n)
        ztrue = zobs + zerr * self.Cinv(c)

        return ztrue

        
    def sample(self, n, zobs_min=None, zobs_max=None):
        """Returns (ztrue, zobs, zerr), where all 3 arrays are shape (n,).'
        If n is large, you may want to call sample_parallel() instead."""

        zobs_min = zobs_min if (zobs_min is not None) else (self.zmin - 0.01)
        zobs_max = zobs_max if (zobs_max is not None) else (self.zmax + 0.01)
        mask = np.logical_and(self.zobs_vec >= zobs_min, self.zobs_vec <= zobs_max)
        rng = np.random.default_rng()
        
        # Shape (n,)
        ix = np.nonzero(mask)[0]
        ix = rng.choice(ix, size=n, replace=True)
        zobs = self.zobs_vec[ix]
        zerr = self.zerr_vec[ix]
        ztrue = np.zeros(n)
        pos = 0

        # Compute in batches, to keep memory usage under control.
        while pos < n:
            pos2 = min(n, pos + 10**5)
            ztrue[pos:pos2] = self._sample(zobs[pos:pos2], zerr[pos:pos2])
            pos = pos2
                
        return ztrue, zobs, zerr


    def sample_parallel(self, n, zobs_min=None, zobs_max=None, processes=None):
        """Returns (ztrue, zobs, zerr), where all 3 arrays are shape (n,).
        
        Uses a multiprocessing Pool for speed. If 'processes' is None, then a
        sensible default will be chosen."""

        with utils.Pool(processes) as pool:
            p = pool._processes
            nlist = [ (((i+1)*n)//p) - ((i*n)//p) for i in range(p) ]
            m = pool.starmap_async(self.sample, [(n,zobs_min,zobs_max) for n in nlist])
            result = m.get()  # list of (ztrue, zobs, zerr) triples
            
        # Concatenate results from each task.
        ztrue = np.concatenate([r[0] for r in result])
        zobs = np.concatenate([r[1] for r in result])
        zerr = np.concatenate([r[2] for r in result])

        return ztrue, zobs, zerr


    @classmethod
    def C(cls, x):
        return (1/2) * scipy.special.erf(x / np.sqrt(2.))

    @classmethod
    def Cinv(cls, y):
        return scipy.special.erfinv(2*y) * np.sqrt(2.)


import os
import h5py
import copy
import fitsio
import numpy as np
import pixell.enmap

from . import utils
from . import io_utils


class Catalog:
    def __init__(self, cols=None, name=None, filename=None, size=0):
        r"""Represents a galaxy catalog, with one "row" per galaxy, and user-defined "columns" for RA, DEC, etc.

        A design decision: do we actually need a Catalog class? Or should we phase it out,
        in favor of ``numpy.recarray`` or ``astropy.Table``?

        Constructor args:
          - cols: dictionary (string col_name) -> (1-d numpy array).
            Optional, since you can add columns later with add_column().
          - name: catalog name (optional)
          - filename: filename, if catalog is stored on disk (optional)
          - size: number of galaxies in catalog (optional, can be inferred from array dimensions instead).

        Members:

          - ``self.size`` (integer): Number of galaxies in catalog.
          - ``self.col_names`` (list of strings): List of user-defined columns.
          - ``self.name`` (string or None): Catalog name (optional).
          - ``self.filename`` (string or None): Filename, if Catalog is stored on disk (optional).

        Additionally, for each column name (in ``self.col_names``), the Catalog contains 
        a member with the corresponding name, whose value is a 1-d array of length self.size.

        Column names are user-defined, but here are some frequently-occuring column names
        (note that I always use lower case):

          - ``self.ra_deg``: right ascension (degrees)
          - ``self.dec_deg``: declination (degrees)
          - ``self.z``: redshift
          - ``self.zerr``: redshift error (if photometric)
          - ``self.rmag``: r-band magnitude (and similarly for g-band, z-band, etc.)

        Here are some functions which return Catalogs::

          kszx.sdss.read_galaxies()
          kszx.sdss.read_randoms()
          kszx.desils_lrg.read_galaxies()
          kszx.desils_lrg.read_randoms()
          kszx.Catalog.from_h5()    # static method
        
        Here are some functions which make maps from catalogs::

          kszx.healpix_utils.map_from_catalog()   # catalog -> 2d healpix map
          kszx.pixell_utils.map_from_catalog()    # catalog -> 2d pixell maps
          kszx.grid_points()                      # catalog -> 3d map
        """

        self.size = size
        self.name = name
        self.col_names = []
        self.filename = filename

        if cols is not None:
            for (col_name, col_data) in cols.items():
                self.add_column(col_name, col_data)

    
    def add_column(self, col_name, col_data):        
        r""" Adds a new column to the catalog.

          - col_name (string)
          - col_data (1-d numpy array)
        """
        
        assert isinstance(col_name, str)
        assert len(col_name) > 0
        assert not col_name.startswith('_')
        assert col_name not in [ 'size', 'name', 'filename', 'col_names' ]
        assert col_name not in self.col_names

        col_data = np.asarray(col_data)
        assert col_data.ndim == 1

        if (len(self.col_names) > 0) or (self.size != 0):
            assert len(col_data) == self.size
        else:
            self.size = len(col_data)
        
        self.col_names.append(col_name)
        setattr(self, col_name, col_data)


    def remove_column(self, col_name):
        r"""Removes an existing column from the catalog."""
        
        assert isinstance(col_name, str)
        assert len(col_name) > 0
        assert not col_name.startswith('_')
        assert col_name not in [ 'size', 'name', 'filename', 'col_names' ]
        assert col_name in self.col_names

        self.col_names.remove(col_name)
        delattr(self, col_name)


    def apply_boolean_mask(self, mask, name=None, in_place=True):
        r"""Reduces the size of the Catalog, by applying a caller-specified boolean mask.
        
        This can be used to implement color cuts, redshift cuts, etc.

           - mask (1-d boolean array). Elements should be True for galaxies which are kept,
             False for galaxies which are discarded.
        
           - name (string or None). If specified, the mask fraction will be printed.

           - in_place (booelan). If False, the original Catalog is unmodified, and a
             new Catalog is returned.
        """
        
        if not in_place:
            self = self.shallow_copy()
                
        mask = np.asarray(mask)
        assert mask.dtype == bool
        if mask.shape != (self.size,):
            raise RuntimeError(f'Catalog.apply_boolean_mask(): expected {mask.shape=} to equal {(self.size,)=}')

        new_size = np.sum(mask)

        if name is not None:
            p = 100 * new_size / self.size
            print(f'{name}: {self.size} -> {new_size} [{p:.05}% retained]')
        
        for col_name in self.col_names:
            col = getattr(self, col_name)
            setattr(self, col_name, col[mask])
            assert getattr(self,col_name).shape == (new_size,)

        self.size = new_size
        return self


    def apply_redshift_cut(self, zmin, zmax, in_place=True):
        r"""Reduces the size of the catalog, by keeping objects in redshift range $z_{min} \le z \le z_{max}$.

        This is a special case of ``apply_boolean_mask()``.

           - zmin (float or None). If None, then no lower redshift limit is imposed.
           - zmax (float or None). If None, then no upper redshift limit is imposed.
           - in_place (booelan). If False, the original Catalog is unmodified, and a
             new Catalog is returned.
        """
        
        assert 'z' in self.col_names

        mask1 = (self.z >= zmin) if (zmin is not None) else np.ones(self.size)
        mask2 = (self.z <= zmax) if (zmax is not None) else np.ones(self.size)
        
        mask = np.logical_and(mask1, mask2)
        name = f'Redshift cut: {zmin=} {zmax=}'
        return self.apply_boolean_mask(mask, name=name, in_place=in_place)


    def get_xyz(self, cosmo, zcol_name='z'):
        r"""Returns shape (N,3) array containing galaxy locations in Cartesian coords (observer at origin). 

        The Catalog must define columns named ``ra_deg``, ``dec_deg``, and ``z``.
        The constructor arg is:

           - cosmo (:class:`~kszx.Cosmology`). Used to convert redshifts to distances.

        This function is super useful -- here are some use cases for the ``points`` array that it returns:
        
           - To "grid" the galaxy field, pass the ``points`` array to :func:`kszx.grid_points`.
           - To interpolate a real-space field to the galaxy locations, pass the ``points`` array 
             to :func:`kszx.interpolate_points`
           - To create a bounding box for the galaxy field, pass the ``points`` array to
             the :class:`~kszx.BoundingBox` constructor.
        
        This function is roughly equivalent to::

           # Compute radial distaance from 'z' column of catalog
           chi = kszx.Cosmology.chi(self.z)
        
           # Compute Cartesian coords from 'ra_deg' and 'dec_deg' cols of catalog.
           points = kszx.utils.ra_dec_to_xyz(self.ra_deg, self.dec_deg, r=chi)
           return points
        """

        assert 'ra_deg' in self.col_names
        assert 'dec_deg' in self.col_names
        assert zcol_name in self.col_names

        chi = cosmo.chi(z = getattr(self,zcol_name))
        xyz = utils.ra_dec_to_xyz(self.ra_deg, self.dec_deg, r=chi)
        return xyz
    

    def generate_batches(self, batchsize, verbose=True):
        r"""Splits catalog into subcatalogs no larger than 'batchsize'.

        If ``batchsize`` is None, the catalog will be "split" into a single subcatalog.

        If ``verbose`` is True, and the catalog is split into more than one batch,
        then a progress indicator will be shown.
        """

        nbatches = 1

        if batchsize is not None:
            assert batchsize > 0
            nbatches = (self.size + batchsize - 1) // batchsize

        if nbatches == 1:
            yield self
            return
        
        for b in range(nbatches):
            s = f'subcatalog {b}/{nbatches}'
            if verbose:
                print('    ', s)  # progress indicator
            if self.name is not None:
                s = f'{self.name} [{s}]'

            i = (b * self.size) // nbatches
            j = ((b+1) * self.size) // nbatches
            assert i < j

            # Subcatalogs are constructed by slicing, not "deep copy".
            yield Catalog(
                cols = { k: getattr(self,k)[i:j] for k in self.col_names },
                name = s,
                size = j-i
            )

            
    def shallow_copy(self):
        return Catalog(
            cols = { s: getattr(self,s) for s in self.col_names },
            name = self.name,
            filename = self.filename,
            size = self.size
        )


    @staticmethod
    def concatenate(catalog_list, name=None, destructive=False):
        r"""Returns a new Catalog, obtained by concatenating (merging) multiple Catalogs.

           - catalog_list (sequence of Catalogs).
           - name (string or None). Name of new Catalog (optional).
           - destructive (boolean). If True, then the input Catalogs will be destroyed to save memory
        """
        
        assert len(catalog_list) > 0
        assert all(isinstance(x, Catalog) for x in catalog_list)

        if any(set(g.col_names) != set(catalog_list[0].col_names) for g in catalog_list):
            raise RuntimeError('Catalog.concatenate(): all Catalogs must have the same column names')

        # copy.copy() is necessary if destructive=True
        col_names = copy.copy(catalog_list[0].col_names)
        cols = { }

        for k in col_names:
            cols[k] = np.concatenate([getattr(g,k) for g in catalog_list])
            if destructive:
                for g in catalog_list:
                    g.remove_column(k)
            
        return Catalog(cols=cols, name=name)


    def absorb_columns(self, src_catalog, destructive=False):
        """'Absorbs' all columns of 'src_catalog' into an existing Catalog (equivalent to calling add_column() multiple times).

        If ``destructive=True``, then all columns will be removed from ``src_catalog`` (saving some memory)."""
        
        if (self.size > 0) and (src_catalog.size > 0) and (self.size != src_catalog.size):
            raise RuntimeError('Catalog.absorb_columns(): source and destination Catalogs must have the same size')
        
        for col_name in copy.copy(src_catalog.col_names):
            col_data = getattr(src_catalog, col_name)
            if not destructive:
                col_data = np.copy(col_data)
                
            self.add_column(col_name, col_data)
            
            if destructive:
                src_catalog.remove_column(col_name)


    def make_random_subcatalog(self, new_size, name=None):
        """Returns a new Catalog, obtained by choosing a random subset (with size ``new_size``) of existing Catalog.

        This method can be used to reduce the size of a random catalog. For example::

           # Random catalong with 100x the galaxy density
           rcat_large = kszx.sdss.read_randoms('CMASS_North')                  

           # Random catalong with 10x the galaxy density
           rcat_small = rcat_large.make_random_subcatalog(rcat_large.size/10)
        """

        print(f'make_random_subcatalog: {self.name}: {self.size} -> {new_size}')

        new_size = int(new_size)
        assert new_size > 0
        assert new_size <= self.size
        
        if name is None:
            name = f'subcatalog({new_size=})'

        # This seems to be the fastest way to select 'new_size' random indices in [0,self.size).
        isrc = np.random.permutation(self.size)[:new_size]
        
        cols = { k: getattr(self,k)[isrc] for k in self.col_names }
        return Catalog(cols=cols, name=name, size=new_size)

    
    @staticmethod
    def from_h5(filename):
        r"""Reads FITS file in HDF5 format (written by ``catalog.write_h5()``), and returns a Catalog object."""
        
        print(f'Reading {filename}\n', end='')
        
        with h5py.File(filename, 'r') as f:
            catalog = Catalog(
                cols = { k: np.asarray(v) for k,v in f.items() },
                name = f.attrs.get('name', default=None),
                filename = filename,
                size = f.attrs['size']
            )

        catalog._announce_file_read()
        return catalog


    def write_h5(self, filename):
        r"""Writes a Catalog to disk, in an HDF5 file format (readable with ``Catalog.read_h5()``)."""
        
        io_utils.mkdir_containing(filename)
        print(f'Writing {filename}\n', end='')
        
        with h5py.File(filename, 'w') as f:
            f.attrs.create('size', self.size)
            
            if self.name is not None:
                f.attrs.create('name', self.name)
                
            for col_name in self.col_names:
                col = getattr(self, col_name)
                assert col.shape == (self.size,)
                f.create_dataset(col_name, data=col)

        self._announce_file_write(filename)

        
    @staticmethod
    def from_fits(filename, col_name_pairs, name=None):
        """Reads FITS file in SDSS/DESILS format, and returns a Catalog object.
        
        The ``col_name_pairs`` arg should be a list of pairs (col_name, fits_col_name).
        See sdss.py and desils_lrg.py for examples."""

        if name is None:
            name = os.path.basename(filename)
            
        print(f'Reading {filename}\n', end='')
        catalog = Catalog(name=name, filename=filename)
        
        with fitsio.FITS(filename) as f:
            for (col_name, fits_col_name) in col_name_pairs:
                catalog.add_column(col_name, f[1].read(fits_col_name))
                
        catalog._announce_file_read()
        return catalog


    @staticmethod
    def from_text_file(filename, col_names, name=None):
        """Reads text file (one row per catalog object) and returns a Catlog object.
        The ``col_names`` arg should be a list of column names. See sdss.py for examples."""

        if name is None:
            name = os.path.basename(filename)
            
        # Note: we copy the transposed array, to get contiguous 1-d arrays.
        print(f'Reading {filename}\n', end='')
        cols = [ np.copy(x) for x in np.loadtxt(filename).T ]

        if len(cols) != len(col_names):
            raise RuntimeError(f'{filename}: file has {len(cols)} columns, {len(col_names)} cols were expected')

        cols = dict(((k,v) for (k,v) in zip(col_names,cols) if k is not None))
        catalog = Catalog(cols, name=name, filename=filename)

        catalog._announce_file_read()
        return catalog


    def show(self):
        """Prints a summary of the Catalog to stdout, with one line per column showing min/mean/max."""
        
        s1 = '' if (self.name is None) else f'name={self.name}, '
        s2 = '' if (self.filename is None) else f', filename={self.filename}'        
        print(f'Catalog({s1}size={self.size}{s2})')

        for k in sorted(self.col_names):
            v = getattr(self, k)
            s = [ ]

            # Numpy operations min/mean/max fail for some dtypes (e.g. uint1).
            try: s.append(f'min={np.min(v)}')
            except: pass
            
            try: s.append(f'mean={np.mean(v)}')
            except: pass
            
            try: s.append(f'max={np.max(v)}')
            except: pass

            # Should always succeed
            s.append(f'dtype={v.dtype}')
            
            print(f"    {k}: {', '.join(s)}")
        
    
    def _announce_file_read(self):
        s1 = '' if (self.name is None) else f'{self.name}: '
        s2 = '' if (self.filename is None) else f' from {self.filename}'
        print(f'{s1}Read {self.size} galaxies{s2}, columns {self.col_names}')

    def _announce_file_write(self, filename):
        s = '' if (self.name is None) else f'{self.name}: '
        print(f'{s}Wrote {self.size} galaxies to {filename}, columns {self.col_names}')
    
    def __str__(self):
        s1 = '' if (self.name is None) else f'name={self.name}, '
        s2 = '' if (self.filename is None) else f', filename={self.filename}'
        return f'Catalog({s1}size={self.size}, col_names={self.col_names}{s2})'



"""
The ``kszx.desi`` module contains functions for downloading/parsing DESI data products.

References:
   - https://data.desi.lbl.gov/doc/releases/dr1/
   - https://data.desi.lbl.gov/doc/organization/
   - https://desidatamodel.readthedocs.io/en/latest/
   - https://github.com/desihub/tutorials
   - https://data.desi.lbl.gov/public/dr1/survey/catalogs/dr1/LSS/iron/LSScats/v1.5/
"""


import os
import fitsio

from . import Catalog
from . import io_utils


def read_galaxies(survey, dr, download=False):
    r"""Reads DESI galaxy catalog, and returns a Catalog object.
    
    DESI catalogs are restricted to an appropriate redshift range, i.e. there is no need
    to apply your own redshift cut after calling this function (unlike SDSS or DESILS-LRG).

    Function arguments:

      - ``survey`` (str): one of the following:
    
           - 'LRG_NGC', 'LRG_SGC'
           - 'QSO_NGC', 'QSO_SGC'
           - 'BGS_ANY_NGC', 'BGS_ANY_SGC',
           - 'BGS_BRIGHT_NGC', 'BGS_BRIGHT_SGC',
           - 'BGS_BRIGHT-21.5_NGC', 'BGS_BRIGHT-21.5_SGC',
           - 'LRG+ELG_LOPnotqso_NGC', 'LRG+ELG_LOPnotqso_SGC'

      - ``dr`` (int): currently only dr=1 is supported

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
      ra_deg, dec_deg, z,    # sky location, redshift
      weight
      wfkp                   # FKP weight
      wcp, wzf, wsys,        # systematic weights

    Example usage::

     # kszx.desi.read_galaxies() returns a kszx.Catalog.
     gcat = kszx.desi.read_galaxies('LRG_NGC', 1)
    """

    filename = _galaxy_filename(survey, dr, download, dlfunc='kszx.desi.read_galaxies')
    gcat = read_fits_catalog(filename, name=survey)
    return gcat


def read_randoms(survey, dr, download=False, nfiles=None):
    r"""Reads DESI random catalog, and returns a Catalog object.

    DESI catalogs are restricted to an appropriate redshift range, i.e. there is no need
    to apply your own redshift cut after calling this function (unlike SDSS or DESILS-LRG).

    Function arguments:

      - ``survey`` (str): one of the following:
    
           - 'LRG_NGC', 'LRG_SGC'
           - 'QSO_NGC', 'QSO_SGC'
           - 'BGS_ANY_NGC', 'BGS_ANY_SGC',
           - 'BGS_BRIGHT_NGC', 'BGS_BRIGHT_SGC',
           - 'BGS_BRIGHT-21.5_NGC', 'BGS_BRIGHT-21.5_SGC',
           - 'LRG+ELG_LOPnotqso_NGC', 'LRG+ELG_LOPnotqso_SGC'

      - ``dr`` (int): currently only dr=1 is supported.

      - ``nfiles`` (int or None): number of random catalog files
        to read. By default, all 18 files are read, which gives
        a random catalog which is ~100 times larger than the galaxy
        catalog.

    Returns a :class:`kszx.Catalog` object, with the following columns::
     
      ra_deg, dec_deg, z,    # sky location, redshift
      weight
      wfkp                   # FKP weight
      wcp, wzf, wsys,        # systematic weights

    Example usage::

     # kszx.desi.read_randoms() returns a kszx.Catalog.
     rcat = kszx.desi.read_randoms('LRG_NGC', 1)
    """
    
    filenames = _random_filenames(survey, dr, download, nfiles, dlfunc='kszx.desi.read_randoms')
    catalogs = [ read_fits_catalog(f) for f in filenames ]
    rcat = Catalog.concatenate(catalogs, name=f'{survey} randoms', destructive=True)
    return rcat


def download(survey, dr, nrfiles):
    r"""Downloads DESI data products (galaxies, randoms) for a given survey.

    Can be called from command line: ``python -m kszx download_desi``.

    Function arguments:

      - ``survey`` (str): one of the following:
    
           - 'LRG_NGC', 'LRG_SGC'
           - 'QSO_NGC', 'QSO_SGC'
           - 'BGS_ANY_NGC', 'BGS_ANY_SGC',
           - 'BGS_BRIGHT_NGC', 'BGS_BRIGHT_SGC',
           - 'BGS_BRIGHT-21.5_NGC', 'BGS_BRIGHT-21.5_SGC',
           - 'LRG+ELG_LOPnotqso_NGC', 'LRG+ELG_LOPnotqso_SGC'

      - ``dr`` (int): currently only dr=1 is supported.

      - ``nrfiles`` (int or None): number of random catalog files
        to read. By default, all 18 files are read, which gives
        a random catalog which is ~100 times larger than the galaxy
        catalog.
    """
    
    _galaxy_filename(survey, dr, download=True)
    _random_filenames(survey, dr, download=True, nfiles=nrfiles)


####################################################################################################


def read_fits_catalog(filename, name=None):
    r"""Reads FITS file in SDSS catalog format. 

    Intended as a helper for read_galaxies() or read_randoms(), but may be useful elsewhere.
    
    Function arguments:

      - ``filename`` (string): should end in ``.fits``.
      - ``name`` (str, optional): name of Catalog, passed to :class:`kszx.Catalog` constructor.

    Returns a :class:`kszx.Catalog` object.
    """

    print(f'Reading {filename}\n', end='')
    catalog = Catalog(name=name, filename=filename)

    with fitsio.FITS(filename) as f:
        
        # Empirically, galaxy/random FITS files contain the following columns:
        #
        #     RA                  f8  
        #     DEC                 f8  
        #     NTILE               i8  
        #     PHOTSYS             S1  
        #     FRAC_TLOBS_TILES    f8
        #     Z                   f8  
        #     WEIGHT              f8  
        #     WEIGHT_SYS          f8  
        #     WEIGHT_COMP         f8  
        #     WEIGHT_ZFAIL        f8  
        #     TARGETID_DATA       i8  
        #     NX                  f8  
        #     WEIGHT_FKP          f8  
        
        catalog.add_column('ra_deg', f[1].read('RA'))
        catalog.add_column('dec_deg', f[1].read('DEC'))
        catalog.add_column('z', f[1].read('Z'))
        catalog.add_column('weight', f[1].read('WEIGHT'))
        catalog.add_column('wfkp', f[1].read('WEIGHT_FKP'))
        catalog.add_column('wcp', f[1].read('WEIGHT_COMP'))   # FIXME is this the same as SDSS WEIGHT_CP?
        catalog.add_column('wzf', f[1].read('WEIGHT_ZFAIL'))  # FIXME is this the same as SDSS WEIGHT_NOZ?
        catalog.add_column('wsys', f[1].read('WEIGHT_SYS'))

    catalog._announce_file_read()
    return catalog

    
def _check_survey(survey, dr):
    """Check that 'survey' is valid, and return its standard capitalization."""
    
    survey_list = [ 
        'LRG_NGC', 'LRG_SGC',
        'QSO_NGC', 'QSO_SGC',
        'BGS_ANY_NGC', 'BGS_ANY_SGC',
        'BGS_BRIGHT_NGC', 'BGS_BRIGHT_SGC',
        'BGS_BRIGHT-21.5_NGC', 'BGS_BRIGHT-21.5_SGC',
        'LRG+ELG_LOPnotqso_NGC', 'LRG+ELG_LOPnotqso_SGC'
    ]
    
    if dr != 1:
        raise RuntimeError(f"DESI {dr=} not supported (currently only dr=1 is supported)")
    
    for s in survey_list:
        if survey.upper() == s.upper():
            return s

    raise RuntimeError(f"DESI survey '{survey}' not recognized (must be one of: {survey_list})")


def _desi_path(relpath, dr, download, dlfunc=None):
    """Intended to be called through wrapper: _galaxy_filename(), _random_filenames(), etc.

    Local filename is:
       {desi_base_dir}/dr1/{relpath}
    
    Remote URL is:
       https://data.desi.lbl.gov/public/dr1/survey/catalogs/dr1/LSS/iron/LSScats/{relpath}

    Note that 'relpath' will begin with v1.5/.
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """

    assert dr == 1   # for now
    desi_base_dir = os.path.join(io_utils.get_data_dir(), 'desi')
    abspath = os.path.join(desi_base_dir, f'dr{dr}', relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://data.desi.lbl.gov/public/dr1/survey/catalogs/dr1/LSS/iron/LSScats/{relpath}'
        io_utils.wget(abspath, url)  # creates directories if needed, asserts os.path.exists(abspath) after download

    return abspath
    

def _galaxy_filename(survey, dr, download=False, dlfunc=None):
    s = _check_survey(survey, dr)
    return _desi_path(f'v1.5/{s}_clustering.dat.fits', dr, download, dlfunc)


def _random_filenames(survey, dr, download=False, nfiles=None, dlfunc=None):
    if nfiles is None:
        nfiles = 18  # default
        
    s = _check_survey(survey, dr)
    return [ _desi_path(f'v1.5/{s}_{i}_clustering.ran.fits', dr, download, dlfunc) for i in range(nfiles) ]



# Reference: https://github.com/pybind/python_example/blob/master/setup.py

from pybind11.setup_helpers import Pybind11Extension, build_ext
from setuptools import setup

cpp_source_files = [
    "cpp/cpp_kernels.cpp",
    "cpp/cic.cpp",
    "cpp/cubic.cpp",
    "cpp/estimate_power_spectrum.cpp",
    "cpp/kbin_average.cpp"
]

ext_module = Pybind11Extension("kszx.cpp_kernels", cpp_source_files, extra_compile_args = ['-O3'])

setup(
    name = 'kszx',
    version = '0.0.1',
    packages = [ 'kszx', 'kszx.tests' ],
    ext_modules = [ ext_module ],
    cmdclass = {"build_ext": build_ext},
    zip_safe = False,
    python_requires = ">=3.11"
)



"""Not much here for now -- just some code to download DR10 randoms. More to come later."""

import os
from . import io_utils


def download_randoms(dr):
    for i in range(20):
        _desils_path(f'randoms/randoms-1-{i}.fits', dr, download=True)


def download_south_randoms(dr):
    for i in range(20):
        _desils_path(f'south/randoms/randoms-south-1-{i}.fits', dr, download=True)

        
def _desils_path(relpath, dr, download=False, dlfunc=None):
    """Example: _desils_path('randoms/randoms-1-0.fits', dr=10).
    
    Intended to be called through wrapper, e.g. _random_filename().
    
    Here and in other parts of kszx, the 'dlfunc' argument gives the name of a transitive caller that
    expects the file to be present, and has a 'download=False' optional argument. This information is
    only used when generating exception-text (to tell the user how to download the file).
    """
    
    assert dr == 10    # placeholder for future expansion
    relpath = os.path.join('dr10', relpath)

    desils_base_dir = os.path.join(io_utils.get_data_dir(), 'desils')
    abspath = os.path.join(desils_base_dir, relpath)

    if io_utils.do_download(abspath, download, dlfunc):
        url = f'https://portal.nersc.gov/cfs/cosmo/data/legacysurvey/{relpath}'
        io_utils.wget(abspath, url)   # calls assert os.path.exists(...) after downloading
    
    return abspath
    



import numpy as np

from .Box import Box
from .Catalog import Catalog
from .Cosmology import Cosmology
from .CatalogGridder import CatalogGridder

from . import utils
from . import core


class KszPSE:
    def __init__(self, box, cosmo, randcat, kbin_edges, surr_ngal_mean, surr_ngal_rms, surr_bg, rweights=None, nksz=0, ksz_rweights=None, ksz_bv=None, ksz_tcmb_realization=None, ztrue_col='z', zobs_col='z', deltac=1.68, kernel='cubic', surr_ic_nbins=1, use_dc=False, spin0_hack=False):
        r"""KszPSE ("KSZ power spectrum estimator"): a high-level pipeline class for $P_{gg}$, $P_{gv}$, and $P_{vv}$.

        Features:

          - Processes multiple velocity reconstructions efficiently in parallel -- for example,
            separate velocity reconstructions for 90 and 150 GHz. This feature can also be used
            to efficiently explore multiple possibilities for the foreground mask, CMB filtering,
            etc, by defining more velocity reconstructions.

          - Runs "surrogate" sims (see overleaf) to characterize the survey window function,
            and assign error bars to power spectra.
 
          - Surrogate sims allow $f_{NL} \ne 0$, to characterize effect of $f_{NL}$ on windowed power 
            spectra, down to arbitrarily low $k$.

          - Velocity reconstruction noise is included in surrogate sims via a bootstrap procedure,
            using the observed CMB realization. This automatically incorporates noise inhomogeneity
            and "striping", and captures correlations e.g. between 90 and 150 GHz.

          - The galaxy catalog can be spectroscopic or photometric (via the ``ztrue_col`` and 
            ``zobs_col`` constructor args). Surrogate sims will capture the effect of photo-z errors.

          - The windowed power spectra $P_{gg}$, $P_{gv}$, $P_{vv}$ use a normalization which
            should be approximately correct. The normalization is an ansatz which is imperfect,
            especially on large scales, so surrogate sims should still be used to compare power
            spetra to models. (However, it's convenient to have a normalization which is not
            "off" by orders of magnitude!) Eventually, we'll implement precise window function
            deconvolution.
        
            Power spectra also include "compensation factors" (see :func:`~kszx.apply_kernel_compensation()`)
            to mitigate interpolation/gridding biases at high $k$.

          - We currently assume only one galaxy field (and only one random field) for simplicity,
            whereas we allow multiple velocity reconstructions. This could be generalized if needed.

        Usage:

          - Construct a KszPSE object (see constructor syntax below).

          - To evaluate power spectra ($P_{gg}$, $P_{gv}$, or $P_{vv}$) on **data or a mock**, 
            call :meth:`KszPSE.eval_pk()` with a galaxy catalog (``gcat``), and filtered CMB 
            temperatures at the galaxy locations (``ksz_tmcb``).

          - To evaluate power spectra on a **surrogate simulation**, call 
            :meth:`KszPSE.simulate_surrogate()` to make the simulation, then call
            :meth:`KszPSE.eval_pk_surrogate()` to estimate power spectra.
        
        Note: Surrogates are factored into two functions (``simulate_surrogate()``and
        ``eval_pk_surrogate()`` so that the caller can put filtering logic in between, by
        operating directly on the ``self.Sg_coeffs``, ``self.Sv_noise``, ``Sv_signal``,
        and ``self.dSg_dfNL`` arrays (see :meth:`KszPSE.simulate_surrogate()` docstring).
        An example of filtering logic is subtracting the mean $\hat v_r$ in redshift bins,
        in order to mitigate foregrounds.
        
        An example notebook where KszPSE is used (but not until later in the notebook):

           https://github.com/kmsmith137/kszx_notebooks/blob/main/05_sdss_pipeline/05_exploratory_plots.ipynb

        The KszPSE computes power spectra involving a galaxy density field $\rho_g$,
        one or more kSZ velocity reconstructions $\hat v_r$, and surrogate fields $S_g, S_v$.
        Definitions of these fields are given in the overleaf, and can be summarized as follows:

        $$\begin{align}
        \rho_g(x) &= \bigg( \sum_{i\in \rm gal} W_i^L \, \delta^3(x-x_i) \bigg) - \frac{N_g}{N_r} \bigg( \sum_{j\in \rm rand} W_j^L \, \delta^3(x-x_j) \bigg) \\
        \hat v_r(x) &= \sum_{i\in \rm gal} W_i^S \, \tilde T(\theta_i) \, \delta^3(x-x_i) \\
        S_g(x) &= \sum_{j\in \rm rand} \frac{N_g}{N_r} W_j^L \big( b_j^G \delta_m(x_j) + \eta_j \big) \delta^3(x-x_j) \\
        S_v(x) &= \sum_{j\in\rm rand} \bigg( \frac{N_g}{N_r} W_j^S b_j^v v_r(x_j) + M_j W_j^S \tilde T(\theta_j) \bigg)  \delta^3(x-x_j)
        \end{align}$$
        For more details, and a formal specification of what KszPSE computes, see ":ref:`ksz_pse_details`" 
        in the sphinx docs.

        Constructor args:

          - ``box`` (:class:`~kszx.Box`): defines pixel size, bounding box size, and location of observer.

          - ``cosmo`` (:class:`~kszx.Cosmology`).

          - ``randcat`` (:class:`~kszx.Catalog`): random catalog, defines survey footprint and redshift
            distribution. The randcat must contain columns ``ra_deg`` and ``dec_deg``.

            **Note:** By default, the galaxy survey is assumed spectroscopic, with a single redshift
            column ``z``. If the catalog is photometric (or if the redshift column is not named ``z``),
            then you'll want to specify the ``ztrue_col`` and ``zobs_col`` constructor args (see below).

          - ``kbin_edges`` (array): 1-d array of length (nkbins+1) defining $k$-bin endpoints for
            power spectrum estimation. The i-th bin covers k-range ``kbin_edges[i] <= i < kbin_edges[i+1]``.

          - ``surr_ngal_mean`` and ``surr_ngal_rms`` (float): In the surrogate sims, I decided to allow $N_{\rm gal}$
            to vary from one surrogate sim to the next. (The idea is to make the surrogate sims more similar to
            mocks, where $N_{\rm gal}$ varies between mocks. Indeed, I find that allowing $N_{\rm gal}$ to vary
            in the surrogate sims does improve overall agreement with mocks.)

            In each surrogate sim, $N_{\rm gal}$ is a Gaussian random variable with mean/rms given by
            the ``surr_ngal_mean`` and ``surr_ngal_rms`` constructor args. If mocks are available, then 
            one way to get sensible values for these arguments is to use the mean/variance in the mocks.
            As a simple placeholder, you could also take ``surr_ngal_rms=0`` (to disable varying $N_{\rm gal}$
            entirely) or ``surr_ngal_rms = sqrt(surr_ngal_mean)`` (Poisson statistics).

          - ``surr_bg``: Galaxy bias $b_g$ used in surrogate sims. Can be specified as either:

               1. an array of length ``randcat.size``, to represent an arbitrary per-object $b_g$.
               2. a callable function $z \rightarrow f(z)$, if $b_g$ only depends on $z$.
               3. a scalar, if $b_g$ is the same for all galaxies.

          - ``rweights`` (optional): Galaxy weighting $W_i^L$ used for the large-scale galaxy field $\delta_g(x)$
            (see equations earlier in this docstring, or ":ref:`ksz_pse_details`" in the sphinx docs).
            Can be specified as either:

               1. an array of length ``randcat.size``, to represent an arbitrary per-object $W_i^L$.
               2. a callable function $z \rightarrow f(z)$, if $W_i^L$ only depends on $z$.
               3. a scalar, if $W_i^L$ is the same for all galaxies.
               4. None (equivalent to ``rweights=1.0``).

            **Note:** there is also a galaxy weighting $W_i^S$ used in the kSZ velocity reconstruction $\hat v_r(x)$.
            The weighting $W_i^S$ is specified as the ``ksz_rweights`` constructor arg, see below.
            The weightings $W_i^L$ and $W_i^S$ are independent.

          - ``nksz`` (integer): Number of kSZ velocity reconstruction fields $\hat v_r(x)$.
            For example, to separate velocity reconstructions from 90 and 150 GHz, use ``nksz=2``.
            If ``nksz`` is zero or unspecified, then the KszPSE will estimate the power spectrum of 
            a galaxy field $\delta_g$, but no $\hat v_r$ fields.

          - ``ksz_rweights`` (optional): Galaxy weighting $W_i^S$ used for the kSZ velocity reconstruction $\hat v_r(x)$.
            Can be specified as either:

              1. an array of length ``randcat.size``, to represent an arbitrary per-object $W_i^S$.
              2. a callable function $z \rightarrow f(z)$, if $W_i^S$ only depends on $z$.
              3. a scalar, if $W_i^S$ is the same for all objects.
              4. None (equivalent to ``ksz_rweights=1``).
              5. a length-``nksz`` list (or iterable) of any of 1-4, if $W_i^S$ is not the same for all
                 velocity reconstructions being processed.

          - ``ksz_bv``: Per-object KSZ velocity bias $b_v^j$, defined by the equation 
            $\tilde T(\theta_j) = b_v^j v_r(x_j) + (\mbox{noise})$. This is only used to assign a normalization
            to power spectra $P_{gv}$ and $P_{vv}$. Can be specified as either:

              1. an array of length ``randcat.size``, to represent an arbitrary per-object $b_v$.
              2. a callable function $z \rightarrow f(z)$, if $b_v$ only depends on $z$.
              3. a scalar, if $b_v$ is the same for all galaxies.
              4. a length-``nksz`` list (or iterable) of any of 1-3, if $b_v$ is not the same for all
                 velocity reconstructions being processed. (This will usually be the case, see next paragraph.)

            The fiducial bias $b_v$ will depend on a choice of fiducial $P_{ge}(k)$, as well as the filter
            applied to $T_{CMB}$. (Since different velocity reconstructions use different CMB filters, $b_v$
            will usually depend on the velocity reconstruction.)

            One reasonable way of initializing $b_v$ is to use the following approximate expression
            from the overleaf:
            $$\begin{align}
            b_j^v &\approx B_v(z_j) \, W_{\rm CMB}(\theta_j) \\
            B_v(\chi) &\equiv \frac{K(\chi)}{\chi^2} \int \frac{d^2L}{(2\pi)^2} \, b_L F_L \, P_{ge}^{\rm true}(k,\chi)_{k=L/\chi}
            \end{align}$$

          - ``ksz_tcmb_realization``: 2-d array of shape ``(nksz, randcat.size)``. This is the filtered CMB $\tilde T$, 
            evaluated at locations of the randoms. This is used when making surrogate sims, to generate bootstrap realizations
            of the reconstruction noise.

            Note that each velocity reconstruction can use a different CMB filter. This is why $\tilde T$ is a 2-d
            array, indexed by ``0 <= i < nksz``, in addition to an index ``0 <= j < randcat.size``.

          - ``ztrue_col`` and ``zobs_col`` (strings): In general, the KszPSE allows the galaxy catalog to be
            photometric, with distinct redshifts $z_{\rm true}, z_{\rm obs}$. A spectroscopic catalog is the
            special case $z_{\rm true} = z_{\rm obs}$.

            The ``ztrue_col`` and ``zobs_col`` args are the names of the $z_{\rm true}, z_{\rm obs}$ columns
            in ``randcat``. The defaults (``ztrue_col = zobs_col = 'z'``) correspond to a spectroscopic catalog,
            with a column ``'z'`` containing redshifts. If the catalog is photometric (or if the redshift column
            is not ``z``), then you'll want to use something different.

          - ``deltac`` (float, default=1.68): This parameter is only used in surrogate sims with $f_{NL} \ne 0$, 
            to compute non-Gaussian bias from Gaussian bias, using $b_{ng} = delta_c (b_g - 1)$.

          - ``kernel`` (string, default=``cubic``): Interpolation kernel passed to ``kszx.interpolate_points()``.
            when simulating the surrogate field. Currently ``cic`` and ``cubic`` are implemented
            (will define more options later).

          - ``surr_ic_nbins`` (boolean): If surr_ic_nbins > 0, then the surrogates will satisfy a global
            "integral constraint" (gal-rand) = 0 in bins.

          - ``spin0_hack`` (booean): Uses a spin-0 estimator for velocity reconstruction (instead of spin-1).
            I'll replace this later by something better!

          - ``use_dc`` (boolean): if False (the default), then the k=0 mode will not be used,
            even if the lowest bin includes k=0.
        """
        
        assert isinstance(box, Box)
        assert isinstance(cosmo, Cosmology)
        assert isinstance(randcat, Catalog)
        assert kernel in [ 'cic', 'cubic' ]  # more kernels will be implemented soon!

        assert randcat.size >= 1000
        assert 'ra_deg' in randcat.col_names
        assert 'dec_deg' in randcat.col_names

        assert surr_ngal_mean > 0
        assert surr_ngal_rms < 5*surr_ngal_mean
        assert surr_ngal_mean + 3.01*surr_ngal_rms <= randcat.size
        assert 0 <= nksz <= 50
        
        self.box = box
        self.cosmo = cosmo
        self.randcat = randcat
        self.kbin_edges = core._check_kbin_edges(box, kbin_edges, use_dc)
        self.nkbins = len(kbin_edges)-1
        self.nksz = nksz
        self.surr_ngal_mean = surr_ngal_mean
        self.surr_ngal_rms = surr_ngal_rms
        self.nrand = randcat.size
        self.ztrue_col = ztrue_col
        self.zobs_col = zobs_col
        self.deltac = deltac
        self.kernel = kernel
        self.surr_ic_nbins = surr_ic_nbins
        self.spin0_hack = spin0_hack
        self.use_dc = use_dc

        ztrue = self._get_zcol(randcat, 'ztrue_col', ztrue_col)
        zobs = self._get_zcol(randcat, 'zobs_col', zobs_col)

        # Small optimization here: if (ztrue is zobs), then don't call ra_dec_to_xyz() twice.
        self.rcat_xyz_true = utils.ra_dec_to_xyz(randcat.ra_deg, randcat.dec_deg, r=cosmo.chi(z=ztrue))
        self.rcat_xyz_obs = utils.ra_dec_to_xyz(randcat.ra_deg, randcat.dec_deg, r=cosmo.chi(z=zobs)) if (ztrue is not zobs) else self.rcat_xyz_true
        
        fname = 'KszPSE.__init__()'
        nrand = self.nrand

        # 1-d arrays of length self.nrand
        self.surr_bg = self._parse_gal_arg(surr_bg, fname, 'surr_bg', nrand, z=ztrue, non_negative=True, allow_none=False)
        self.rweights = self._parse_gal_arg(rweights, fname, 'rweights', nrand, z=zobs, non_negative=True, allow_none=True)

        # Length-nksz lists of (1-d array or None)
        self.ksz_bv = self._parse_ksz_arg(ksz_bv, fname, 'ksz_bv', nrand, z=ztrue, allow_none=False)
        self.ksz_rweights = self._parse_ksz_arg(ksz_rweights, fname, 'ksz_rweights', nrand, z=ztrue, allow_none=True)
        self.ksz_tcmb_realization = self._parse_tcmb_arg(ksz_tcmb_realization, fname, 'ksz_tcmb_realization', nrand)
        assert len(self.ksz_bv) ==  len(self.ksz_rweights) == len(self.ksz_tcmb_realization) == self.nksz
        
        # 1-d arrays of length self.nrand
        self.D = cosmo.D(z=ztrue, z0norm=True)
        self.faH = cosmo.frsd(z=ztrue) * cosmo.H(z=ztrue) / (1+ztrue)
        self.sigma2 = self._integrate_kgrid(self.box, cosmo.Plin_z0(self.box.get_k()))
        self.zobs = zobs   # needed for surr_ic_nbins

        # Check that we have enough randoms to make surrogate fields.
        self.bD_max = np.max(self.surr_bg * self.D)
        self.ngal_min = surr_ngal_mean - 3*surr_ngal_rms  # clamped to 3 sigma
        self.ngal_max = surr_ngal_mean + 3*surr_ngal_rms  # clamped to 3 sigma
        self.nrand_min = int(self.bD_max**2 * self.sigma2 * (self.ngal_max + 10))
        assert self.ngal_min >= 10
        
        if self.nrand < self.nrand_min:
            raise RuntimeError(f'KszPSE: not enough randoms to make surrogate fields! This can be fixed by using'
                               + f' a larger random catalog (nrand={self.nrand}, nrand_min={self.nrand_min}).')

        # pse_rweights = Length (nksz+1) list of 1-d arrays.
        pse_rweights = [ self.rweights ]

        for (w,bv) in zip(self.ksz_rweights, self.ksz_bv):
            pse_rweights += [ (w*bv) if (w is not None) else bv ]

        self.catalog_gridder = CatalogGridder(
            box = box,
            cosmo = cosmo,
            randcat = randcat,
            rweights = pse_rweights,
            nfootprints = nksz + 1,
            zcol_name = zobs_col,   # not ztrue_col
            save_rmaps = [True] + [False]*nksz,
            kernel = self.kernel
        )

    
    def eval_pk(self, gcat, gweights=None, ksz_gweights=None, ksz_bv=None, ksz_tcmb=None, zobs_col=None):
        r"""Computes $P_{gg}$, $P_{gv}$, and $P_{vv}$ from data or a mock (not a surrogate sim).

        Function args:

          - ``gcat`` (:class:`~kszx.Catalog`): galaxy catalog.

          - ``gweights`` (optional). Galaxy weighting $W_i^L$ used for the large-scale galaxy field $\delta_g(x)$.
             Can be specified as either:

              1. an array of length ``gcat.size``, to represent an arbitrary per-object $W_i^L$.
              2. a callable function $z \rightarrow f(z)$, if $W_i^L$ only depends on $z$.
              3. a scalar, if $W_i^L$ is the same for all galaxies.
              4. None (equivalent to ``gweights=1.0``).

            The ``gweights`` passed to ``eval_pk()`` should reflect as closely as possible the ``rweights``
            passed to the :class:`~kszx.KszPSE` constructor. For example, if randoms are FKP-weighted, then
            galaxies should also be FKP-weighted. (Warning: even a small mismatch between the weighting of
            galaxies/randoms can produce a large contribution to $P_{gg}$!)

          - ``ksz_gweights`` (optional): Galaxy weighting $W_i^S$ used for the kSZ velocity reconstruction $\hat v_r(x)$.
            Can be specified as either:

              1. an array of length ``randcat.size``, to represent an arbitrary per-object $W_i^S$.
              2. a callable function $z \rightarrow f(z)$, if $W_i^S$ only depends on $z$.
              3. a scalar, if $W_i^S$ is the same for all objects.
              4. None (equivalent to ``ksz_rweights=1``).
              5. a length-``nksz`` list (or iterable) of any of 1-4, if $W_i^S$ is not the same for all
                 velocity reconstructions being processed.

            The ``ksz_gweights`` passed to ``eval_pk()`` should reflect as closely as possible the 
            ``ksz_rweights`` passed to the :class:`~kszx.KszPSE` constructor.
    
          - ``ksz_bv``: Per-object KSZ velocity bias $b_v^i$. For more discsusion, and some key equations
            showing how to approximate $b_v$, see the ``ksz_bv`` argument in the :class:`~kszx.KszPSE`
            constructor docstring. Can be specified as either:

              1. an array of length ``gcat.size``, to represent an arbitrary per-object $b_v$.
              2. a callable function $z \rightarrow f(z)$, if $b_v$ only depends on $z$.
              3. a scalar, if $b_v$ is the same for all galaxies.
              4. a length-``nksz`` list (or iterable) of any of 1-3, if $b_v$ is not the same for all
                 velocity reconstructions bing processed. (This will usually be the case.)

            The ``ksz_bv`` argument passed to ``eval_pk()`` should reflect as closely as possible the 
            ``ksz_bv`` argument passed to the :class:`~kszx.KszPSE` constructor. (Same filtering, etc.)

          - ``ksz_tcmb``: 2-d array of shape ``(nksz, gcat.size)``.
            Note that each velocity reconstruction can use a different CMB filter. This is why $\tilde T$ is a 2-d
            array, indexed by ``0 <= i < nksz``, in addition to an index ``0 <= j < gcat.size``.

            The ``ksz_tcmb`` argument passed to ``eval_pk()`` should reflect as closely as possible the 
            ``ksz_tcmb_realization`` argument passed to the :class:`~kszx.KszPSE` constructor. (Same filtering, etc.)
        
          - ``zobs_col`` (string): name of column containing observed redshifts in ``gcat``.
            By default, we use the value ``zcol_obs`` that was specified when the constructor was called.

        Return value:

          - Array ``pk`` with shape ``(nksz+1, nksz+1, nkbins)``. The first two indices are
            field indices, with ordering (galaxy field, vrec fields). Thus the ``pk`` array
            contains all auto and cross power spectra $P_{gg}$, $P_{gv_i}$, $P_{v_iv_j}$.
        """

        assert isinstance(gcat, Catalog)

        if zobs_col is None:
            zobs_col = self.zobs_col
            
        z = self._get_zcol(gcat, 'zobs_col', zobs_col)
        gcat_xyz = gcat.get_xyz(self.cosmo, zcol_name=zobs_col)

        # Parse the 'gweights', 'ksz_gweights', 'ksz_bv', and 'ksz_tcmb' args.
        # (Same parsing logic as 'rweights', 'ksz_rweights', 'ksz_bv', and 'ksz_tcmb_realization' in __init__().)

        fname = 'KszPSE.eval_pk()'
        gweights = self._parse_gal_arg(gweights, fname, 'gweights', gcat.size, z, non_negative=True, allow_none=True)
        ksz_bv = self._parse_ksz_arg(ksz_bv, fname, 'ksz_bv', gcat.size, z, allow_none=False)
        ksz_gweights = self._parse_ksz_arg(ksz_gweights, fname, 'ksz_gweights', gcat.size, z, allow_none=True)
        ksz_tcmb = self._parse_tcmb_arg(ksz_tcmb, fname, 'ksz_tcmb', gcat.size)
        assert len(ksz_bv) == len(ksz_gweights) == len(ksz_tcmb) == self.nksz

        # Initialize fmaps.

        gweights = gweights if (gweights is not None) else 1.0
        fmaps = [ self.catalog_gridder.grid_density_field(gcat, gweights, 0, zcol_name=zobs_col) ]

        for i in range(self.nksz):
            w, bv, t = ksz_gweights[i], ksz_bv[i], ksz_tcmb[i]
            coeffs = (w*t) if (w is not None) else t
            wsum = np.dot(w,bv) if (w is not None) else np.sum(bv)
            spin = 0 if self.spin0_hack else 1
            fmaps += [ self.catalog_gridder.grid_sampled_field(gcat, coeffs, wsum, i+1, spin=spin, zcol_name=zobs_col) ]

        # FIXME need some sort of CatalogPSE helper function here.
        pk = core.estimate_power_spectrum(self.box, fmaps, self.kbin_edges)
        pk *= np.reshape(self.catalog_gridder.ps_normalization, (self.nksz+1, self.nksz+1, 1))
        return pk
        

    def simulate_surrogate(self):
        r"""Makes a random surrogate simulation, and stores the result in class members (return value is None).
        
        $$\begin{align}
        S_g(x) &= \sum_{j\in\rm rand} S_g^j \delta^3(x-x_j) \\
        S_v(x) &= \sum_{j\in\rm rand} S_v^j \delta^3(x-x_j)
        \end{align}$$

        Initializes the following class members:

          - ``self.surr_ngal`` (integer): value of $N_{\rm gal}$ in the surrogate sim. (Recall that
            the value of $N_{\rm gal}$ in each surrogate sim is a Gaussian random variable. See
            ``surr_ngal_mean`` and ``surr_ngal_rms`` in the :class:`~kszx.KszPSE` constructor
            docstring for more info.)

          - ``self.Sg_coeffs`` (1-d array of length ``self.nrand``): Coefficients $S_g^j$ above.

          - ``self.Sv_noise`` (shape-``(self.nksz, self.nrand)`` array): Coefficients $S_v^j$ above,
            contribution from reconstruction noise only.
        
          - ``self.Sv_signal`` (shape-``(self.nksz, self.nrand)`` array): Coefficients $S_v^j$ above,
            contribution from velocity field only (i.e. no reconstruction noise).

          - ``self.dSg_dfNL`` (1-d array of length ``self.nrand``): Derivative $dS_g^j/df_{NL}$.

        Note: Surrogates are factored into two functions (``simulate_surrogate()``and
        ``eval_pk_surrogate()`` so that the caller can put filtering logic in between, by
        operating directly on the ``self.Sg_coeffs``, ``self.Sv_noise``, ``Sv_signal``,
        and ``self.dSg_dfNL`` arrays. An example of filtering logic is subtracting the
        mean $\hat v_r$ in redshift bins, in order to mitigate foregrounds.
        """

        ngal = self.surr_ngal_mean + (self.surr_ngal_rms * np.random.normal())
        ngal = np.clip(ngal, self.ngal_min, self.ngal_max)
        ngal = int(ngal+0.5)  # round
        self.surr_ngal = ngal
        
        nrand = self.nrand
        assert 0 < ngal <= nrand
                
        delta0 = core.simulate_gaussian_field(self.box, self.cosmo.Plin_z0)
        core.apply_kernel_compensation(self.box, delta0, self.kernel)

        # delta_g(k,z) = (bg + 2 fNL deltac (bg-1) / alpha(k,z)) * delta_m(k,z)
        #              = (bg D(z) + 2 fNL deltac (bg-1) / alpha0(k)) * delta0(k)

        Sg_prefactor = ((ngal/nrand) * self.rweights) if (self.rweights is not None) else (ngal/nrand)
        
        # deltaG = (bg * delta_m) = (bg * D * delta0), evaluated on randcat.
        bD = self.surr_bg * self.D
        delta_G = bD * core.interpolate_points(self.box, delta0, self.rcat_xyz_true, self.kernel, fft=True)
        self.Sg_coeffs = Sg_prefactor * delta_G

        eta_rms = np.sqrt((nrand/ngal) - (bD*bD) * self.sigma2)
        eta = np.random.normal(scale = eta_rms)
        self.Sg_coeffs += Sg_prefactor * eta

        # Add term to deltag:
        #     2 fNL deltac (bg-1) / alpha(k,z) * delta_m(k,z)
        #   = 2 fNL deltac (bg-1) / alpha_z0(k) * delta0(k)    [ factor D(z) cancels ]
        phi0 = core.multiply_kfunc(self.box, delta0, lambda k: 1.0/self.cosmo.alpha_z0(k=k), dc=0)
        phi0 = core.interpolate_points(self.box, phi0, self.rcat_xyz_true, self.kernel, fft=True)
        self.dSg_dfnl = Sg_prefactor * (2 * self.deltac) * (self.surr_bg-1) * phi0

        if self.surr_ic_nbins > 0:
            self.Sg_coeffs[:] = utils.subtract_binned_means(self.Sg_coeffs, self.zobs, self.surr_ic_nbins)
            self.dSg_dfnl[:] = utils.subtract_binned_means(self.dSg_dfnl, self.zobs, self.surr_ic_nbins)
        
        self.Sv_noise = np.zeros((self.nksz, self.nrand))
        self.Sv_signal = np.zeros((self.nksz, self.nrand))
        
        if self.nksz == 0:
            return

        # vr = (faHD/k) delta0, evaluated at rcat_xyz_true.
        vr = core.multiply_kfunc(self.box, delta0, lambda k: 1.0/k, dc=0)
        vr = core.interpolate_points(self.box, vr, self.rcat_xyz_true, self.kernel, fft=True, spin=1)
        vr *= self.faH
        vr *= self.D

        # M = vector with (nrand-ngal) 0s and (ngal) 1s, in random positions.
        # (This way of generating M is a little slow, but I don't think there's a faster way to do
        # it in numpy, and it's not currently a bottleneck.)
        M = np.zeros(nrand)
        M[:ngal] = 1.0
        M = np.random.permutation(M)

        for i in range(self.nksz):
            self.Sv_noise[i,:] = M * self.ksz_tcmb_realization[i]
            self.Sv_signal[i,:] = (ngal/nrand) * self.ksz_bv[i] * vr

            if self.ksz_rweights[i] is not None:
                self.Sv_noise[i,:] *= self.ksz_rweights[i]
                self.Sv_signal[i,:] *= self.ksz_rweights[i]


    def eval_pk_surrogate(self):
        r"""Returns an array of shape (2*nksz+2, 2*nksz+2, nkbins), obtained from the current surrogate sim.

        The first two indices of the array are "field indices" as follows:
            0 = surrogate sim S_g, with fNL=0.
            1 = surrogate dS_g/dfNL
            (2*i+2) = surrogate sim S_v, reconstruction noise contribution only (where 0 <= i < nksz).
            (2*i+3) = surrogate sim S_v, velocity signal contribution only (where 0 <= i < nksz).

        Thus, the returned shape (2*nksz+2, 2*nksz+2, nkbins) array can be used to compute P_gg, P_gv, P_vv
        for all combinations of CMB frequencies, and arbitrary (fNL, bv).

        Reminder: logic for surrogate sims is split between :meth:`simulate_surrogate()` and this method.
        When :meth:`simulate_surrogate()` is called, it initializes coefficient arrays (``self.Sg_coeffs`` etc.)
        which are used here to simulate surrogate fields and estimate power spectra.
        """

        if not hasattr(self, 'Sg_coeffs'):
            raise RuntimeError("Before calling KszPSE.eval_pk_surrogate(), you must call KszPSE.simulate_surrogate()")
        
        Nr = self.nrand
        Ng = self.surr_ngal
        Sg_wsum = ((Ng/Nr) * np.sum(self.rweights)) if (self.rweights is not None) else Ng

        # fmaps = list of Fourier-space maps
        # fmaps[0] = surrogate sim S_g, with fNL=0.
        # fmaps[1] = surrogate dS_g/dfNL
        # fmaps[2*i+2] = surrogate sim S_v, reconstruction noise contribution only (where 0 <= i < nksz).
        # fmaps[2*i+3] = surrogate sim S_v, velocity signal contribution only (where 0 <= i < nksz).
        
        fmaps = [ ]
        fmaps += [ self.catalog_gridder.grid_sampled_field(self.randcat, self.Sg_coeffs, Sg_wsum, 0, zcol_name=self.zobs_col) ]
        fmaps += [ self.catalog_gridder.grid_sampled_field(self.randcat, self.dSg_dfnl, Sg_wsum, 0, zcol_name=self.zobs_col) ]
        
        for i in range(self.nksz):
            w, bv = self.ksz_rweights[i], self.ksz_bv[i]
            Sv_wsum = (Ng/Nr) * (np.dot(w,bv) if (w is not None) else np.sum(bv))
            spin = 0 if self.spin0_hack else 1
            fmaps += [ self.catalog_gridder.grid_sampled_field(self.randcat, self.Sv_noise[i,:], Sv_wsum, i+1, spin=spin, zcol_name=self.zobs_col) ]
            fmaps += [ self.catalog_gridder.grid_sampled_field(self.randcat, self.Sv_signal[i,:], Sv_wsum, i+1, spin=spin, zcol_name=self.zobs_col) ]

        # Unnormalized power spectrum estimates (from CatalogGridder)
        pk = core.estimate_power_spectrum(self.box, fmaps, self.kbin_edges)
        assert pk.shape == (2*self.nksz+2, 2*self.nksz+2, self.nkbins)

        # Index mapping for power spectrum normalization.
        imap = [ (i//2) for i in range(2*self.nksz+2) ]

        # Apply power spectrum normalization
        for i in range(2*self.nksz+2):
            for j in range(2*self.nksz+2):
                pk[i,j,:] *= self.catalog_gridder.ps_normalization[imap[i],imap[j]]

        return pk


    @classmethod
    def reduce_pk(cls, pk, fnl, bv):
        """Reduces the (2*nksz+2, 2*nksz+2, nkbins)  array returned by eval_pk_surrogate() by fixing fNL and bv.

        The input array 'pk' can either be
          - an array of shape (2*nksz+2, 2*nksz+2, nkbins) returned by eval_pk_surrogate(), representing one sim
          - an array of shape (nsurr, 2*nksz+2, 2*nksz+2, nkbins) representing many sims

        The output array has shape
          - (nksz+1, nksz+1, nkbins) in the first case above
          - (nsurr, nksz+1, nksz+1, nkbins) in the second case
        """

        pk = np.asarray(pk)
        
        if not cls._valid_shape_for_reduce_pk(pk):
            raise RuntimeError(f"KszPSE.reduce_pk(): 'pk' argument has invalid shape {pk.shape}")

        nksz = (pk.shape[1]//2) - 1
        coeffs = np.array([fnl] + [bv]*nksz)

        # Reduce axis 0
        if pk.ndim == 3:
            pk = pk[::2] + pk[1::2] * coeffs.reshape((nksz+1,) + (1,)*(pk.ndim-1))

        # Reduce axis 1
        pk = pk[:,::2] + pk[:,1::2] * coeffs.reshape((1,nksz+1) + (1,)*(pk.ndim-2))

        # Reduce axis 2
        if pk.ndim == 4:
            pk = pk[:,:,::2] + pk[:,:,1::2] * coeffs.reshape((1,1,nksz+1) + (1,)*(pk.ndim-3))

        return pk
    
        
    @classmethod
    def _valid_shape_for_reduce_pk(cls, pk):
        if (pk.ndim < 3) or (pk.ndim > 4):
            return False
        if pk.shape[pk.ndim-3] != pk.shape[pk.ndim-2]:
            return False
        if (pk.shape[1] % 2) or (pk.shape[1] < 4):
            return False
        return True
        
        
    ####################################################################################################


    def _get_zcol(self, catalog, zcol_argname, zcol_name):
        """Helper method called by constructor, to check that column exists in catalog."""

        if not isinstance(zcol_name, str):
            raise RuntimeError(f"KszPSE: expected argument {zcol_argname}={zcol_name} to be a string")

        if zcol_name not in catalog.col_names:
            raise RuntimeError(f"KszPSE: catalog does not contain column '{zcol_name}' (specified as argument '{zcol_argname}')")

        z = getattr(catalog, zcol_name)
        zmin = np.min(z)
        
        if zmin <= 0:
            raise RuntimeError(f"KszPSE: expected all redshifts to be positive (got min(catalog.{zcol_name}) = {zmin})")

        return z

    
    def _parse_gal_arg(self, f, funcname, argname, ngal, z, allow_none=False, non_negative=False):
        """Used to parse constructor args 'surr_bg' and 'rweights'.

        The 'f' argument is either:
           1. a callable function z -> f(z)
           2. an array of length ngal
           3. a scalar
           4. None (equivalent to 1.0)

        If 'z' is None, then a function (option #1 above) is not allowed.
        If 'allow_none' is False, then None (option #4 above) is not allowed.

        Returns an array of length ngal.
        """

        if f is None:
            if allow_none:
                return np.full(ngal, 1.0)
            raise RuntimeError(f"{funcname}: '{argname}' argument must be specified")
        
        if callable(f):
            if z is None:
                raise RuntimeError(f"{funcname}: '{argname}' constructor argument is a function -- this is not currently allowed")
            fz = f(z)
            fz_name = f'return value from {argname}()'
            fz = utils.asarray(fz, funcname, fz_name, dtype=float)
            if fz.shape != (ngal,):
                raise RuntimeError(f"{fz_name} has shape {fz.shape}, expected shape {(ngal,)}")
        
        else:
            fz = utils.asarray(f, funcname, argname, dtype=float)
        
            if fz.ndim == 0:
                fz = np.full(ngal, fz)
            elif fz.shape != (ngal,):
                raise RuntimeError(f"{funcname}: expected '{argname}' constructor arg to be either"
                                   + f" shape ({ngal},) or a scalar (got shape {fz.shape})")

        assert fz.shape == (ngal,)
        assert fz.dtype == float

        if non_negative and np.min(fz) < 0:
            raise RuntimeError(f"{funcname}: expected {argname} to be >= 0")

        return fz


    def _parse_ksz_arg(self, f, funcname, argname, ngal, z, allow_none=False, non_negative=False):
        """
        Used to parse constructor args 'ksz_rweights', 'ksz_bv', 'ksz_tcmb_realization'.

        The 'f' argument is either:
           1. a callable function z -> f(z)
           2. an array of length ngal
           3. a scalar
           4. None (equivalent to 1.0)
           5. a length-nksz list (or iterable) of any of 1-4.

        If 'z' is None, then a function (option #1 above) is not allowed.
        If 'allow_none' is False, then None (option #4 above) is not allowed.
        Returns a length-nksz list of (1-d numpy array or None), where 1-d numpy arrays have length ngal.
        """

        if self.nksz == 0:
            # All kSZ-related constructor args must be either None, or an empty iterable.
            if (f is None) or self._is_empty_iterable(f):
                return []
            raise RuntimeError(f"{funcname}: argument '{argname}' was specified, but 'nksz' constructor arg was unspecified (or zero)")

        is_small_array = isinstance(f, np.ndarray) and (f.shape[0] != self.nksz)

        # These cases can be handled by a single call to self._parse_gal_arg().
        if is_small_array or (f is None) or callable(f) or self._is_scalar(f):
            x = self._parse_gal_arg(f, funcname, argname, ngal, z, allow_none=allow_none, non_negative=non_negative)
            return [x] * self.nksz

        try:
            flen = len(f)
        except:
            raise RuntimeError(f"{funcname}: couldn't parse argument '{argname}' (got type {type(f)},"
                               + " expected iterable, function, scalar, or None)")

        if flen != self.nksz:
            raise RuntimeError(f"{funcname}: couldn't parse argument '{argname}' (got length-{flen} iterable non-array,"
                               + f" expected array, function, scalar, None, or iterable with length nksz={self.nksz})")

        # This case can be handled by multiple calls to self._parse_gal_arg().
        return [ self._parse_gal_arg(x, funcname, f'{argname}[{i}]', ngal, z, allow_none=allow_none, non_negative=non_negative) for i,x in enumerate(f) ]

    
    def _parse_tcmb_arg(self, x, funcname, argname, ngal):
        """Used to parse 'ksz_tcmb_realization' constructor arg, or 'ksz_tcmb' arg to eval_pk()."""
        
        if self.nksz == 0:
            if (x is None) or self._is_empty_array(f):
                return np.zeros((self.nksz,ngal), dtype=float)
            raise RuntimeError(f"{funcname}: argument '{argname}' was specified, but 'nksz' constructor arg was unspecified (or zero)")

        x = utils.asarray(x, funcname, argname, dtype=float)

        if x.shape == (self.nksz, ngal):
            return x
        
        raise RuntimeError(f"{funcname}: expected argument '{argname}' to have shape (nksz,ngal)={(self.nksz,ngal)}, got shape {x.shape}")

    
    @staticmethod
    def _is_empty_iterable(f):
        try:
            for _ in f:
                return False
            return True
        except:
            pass
        return False

    
    @staticmethod
    def _is_empty_array(x):
        try:
            if np.asarray(x).size == 0:
                return True
        except:
            pass
        return False
    
    
    @staticmethod
    def _is_scalar(f):
        try:
            f = float(f)
            return True
        except:
            pass
        return False

    
    @staticmethod
    def _integrate_kgrid(box, kgrid):
        """Helper method for constructor. Could be moved somewhere more general (e.g. kszx.core)."""

        assert kgrid.shape == box.fourier_space_shape
    
        w = np.full(box.fourier_space_shape[-1], 2.0)
        w[0] = 1.0
        w[-1] = 2.0 if (box.real_space_shape[-1] % 2) else 1.0
        w = np.reshape(w, (1,)*(box.ndim-1) + (-1,))

        # FIXME could be optimized
        ret = np.sum(w * kgrid) / box.box_volume
        return np.real(ret)
